# -*- coding: utf-8 -*-
"""
controller functions for contributors, versions, configs, and other metadata
"""
import logging
import functools
import utool as ut
from six.moves import range, input, zip
from wbia import constants as const
from wbia.control import accessor_decors, controller_inject
from wbia.algo import Config
from wbia.control.controller_inject import make_ibs_register_decorator

print, print_, profile = ut.inject2(__name__)
logger = logging.getLogger('wbia')


CLASS_INJECT_KEY, register_ibs_method = make_ibs_register_decorator(__name__)


register_api = controller_inject.get_wbia_flask_api(__name__)


@register_ibs_method
@accessor_decors.adder
@register_api('/api/contributor/', methods=['POST'])
def add_contributors(
    ibs,
    tag_list,
    uuid_list=None,
    name_first_list=None,
    name_last_list=None,
    loc_city_list=None,
    loc_state_list=None,
    loc_country_list=None,
    loc_zip_list=None,
    notes_list=None,
):
    r"""
    Adds a list of contributors.

    Returns:
        contributor_id_list (list): contributor rowids

    RESTful:
        Method: POST
        URL:    /api/contributor/
    """
    import datetime

    def _valid_zip(_zip, default='00000'):
        _zip = str(_zip)
        if len(_zip) == 5 and _zip.isdigit():
            return _zip
        return default

    if ut.VERBOSE:
        logger.info('[ibs] adding %d imagesets' % len(tag_list))
    # Add contributors to database
    if name_first_list is None:
        name_first_list = [''] * len(tag_list)
    if name_last_list is None:
        name_last_list = [''] * len(tag_list)
    if loc_city_list is None:
        loc_city_list = [''] * len(tag_list)
    if loc_state_list is None:
        loc_state_list = [''] * len(tag_list)
    if loc_country_list is None:
        loc_country_list = [''] * len(tag_list)
    if loc_zip_list is None:
        loc_zip_list = [''] * len(tag_list)
    if notes_list is None:
        notes_list = [
            'Created %s' % (datetime.datetime.now(),) for _ in range(len(tag_list))
        ]

    loc_zip_list = [_valid_zip(_zip) for _zip in loc_zip_list]

    if uuid_list is None:
        # contributor_rowid_list = ibs.get_contributor_rowid_from_tag(tag_list)
        # uuid_list = ibs.get_contributor_uuid(contributor_rowid_list)
        # uuid_list = ibs.get_contributor_uuid(contributor_rowid_list)
        # uuid_list = [ uuid.uuid4() if uuid_ is None else uuid_ for uuid_ in uuid_list ]
        # DETERMENISTIC UUIDS
        zero_uuid = ut.get_zero_uuid()
        uuid_list = [ut.augment_uuid(zero_uuid, tag) for tag in tag_list]

    colnames = [
        'contributor_uuid',
        'contributor_tag',
        'contributor_name_first',
        'contributor_name_last',
        'contributor_location_city',
        'contributor_location_state',
        'contributor_location_country',
        'contributor_location_zip',
        'contributor_note',
    ]
    params_iter = zip(
        uuid_list,
        tag_list,
        name_first_list,
        name_last_list,
        loc_city_list,
        loc_state_list,
        loc_country_list,
        loc_zip_list,
        notes_list,
    )

    get_rowid_from_superkey = ibs.get_contributor_rowid_from_uuid
    # get_rowid_from_superkey = ibs.get_contributor_rowid_from_tag  # ?? is tag a superkey?
    contributor_id_list = ibs.db.add_cleanly(
        const.CONTRIBUTOR_TABLE, colnames, params_iter, get_rowid_from_superkey
    )
    return contributor_id_list


@register_ibs_method
@accessor_decors.adder
def add_version(ibs, versiontext_list):
    r"""
    Adds an algorithm / actor configuration as a string
    """
    # FIXME: Configs are still handled poorly
    params_iter = ((versiontext,) for versiontext in versiontext_list)
    get_rowid_from_superkey = ibs.get_version_rowid_from_superkey
    versionid_list = ibs.db.add_cleanly(
        const.VERSIONS_TABLE, ('version_text',), params_iter, get_rowid_from_superkey
    )
    return versionid_list


# SETTERS::METADATA


@register_ibs_method
@accessor_decors.setter
# @register_api('/api/metadata/value/', methods=['PUT'])
def set_metadata_value(ibs, metadata_key_list, metadata_value_list, db):
    r"""
    Sets metadata key, value pairs

    RESTful:
        Method: PUT
        URL:    /api/metadata/value/
    """
    db = db[0]  # Unwrap tuple, required by @accessor_decors.setter decorator
    metadata_rowid_list = ibs.get_metadata_rowid_from_metadata_key(metadata_key_list, db)
    id_iter = ((metadata_rowid,) for metadata_rowid in metadata_rowid_list)
    val_list = ((metadata_value,) for metadata_value in metadata_value_list)
    db.set(const.METADATA_TABLE, ('metadata_value',), val_list, id_iter)


@register_ibs_method
def set_database_version(ibs, db, version):
    r"""
    Sets the specified database's version from the controller
    """
    db.set_db_version(version)


# SETTERS::CONTRIBUTORS


@register_ibs_method
# @register_api('/api/config/contributor/rowid/', methods=['PUT'])
def set_config_contributor_rowid(ibs, config_rowid_list, contributor_rowid_list):
    r"""
    Sets the config's contributor rowid

    RESTful:
        Method: PUT
        URL:    /api/config/contributor/rowid/
    """
    id_iter = ((config_rowid,) for config_rowid in config_rowid_list)
    val_list = ((contributor_rowid,) for contributor_rowid in contributor_rowid_list)
    ibs.db.set(const.CONFIG_TABLE, ('contributor_rowid',), val_list, id_iter)


@register_ibs_method
# @register_api('/api/contributor/new/temp/', methods=['POST'])
def add_new_temp_contributor(ibs, user_prompt=False, offset=None, autolocate=False):
    r"""

    RESTful:
        Method: POST
        URL:    /api/contributor/new/temp/
    """
    name_first = ibs.get_dbname()
    name_last = ut.get_computer_name() + ':' + ut.get_user_name() + ':' + ibs.get_dbdir()
    logger.info(
        '[collect_transfer_data] Contributor default first name: %s' % (name_first,)
    )
    logger.info(
        '[collect_transfer_data] Contributor default last name:  %s' % (name_last,)
    )
    if user_prompt:
        name_first = input(
            '\n[collect_transfer_data] Change first name (Enter to use default): '
        )
        name_last = input(
            '\n[collect_transfer_data] Change last name (Enter to use default): '
        )

    if autolocate:
        (
            success,
            location_city,
            location_state,
            location_country,
            location_zip,
        ) = ut.geo_locate()
    else:
        success = False

    if success:
        logger.info(
            '\n[collect_transfer_data] Your location was be determined automatically.'
        )
        logger.info(
            '[collect_transfer_data] Contributor default city: %s' % (location_city,)
        )
        logger.info(
            '[collect_transfer_data] Contributor default state: %s' % (location_state,)
        )
        logger.info(
            '[collect_transfer_data] Contributor default zip: %s' % (location_country,)
        )
        logger.info(
            '[collect_transfer_data] Contributor default country: %s' % (location_zip,)
        )
        if user_prompt:
            location_city = input(
                '\n[collect_transfer_data] Change default location city (Enter to use default): '
            )
            location_state = input(
                '\n[collect_transfer_data] Change default location state (Enter to use default): '
            )
            location_zip = input(
                '\n[collect_transfer_data] Change default location zip (Enter to use default): '
            )
            location_country = input(
                '\n[collect_transfer_data] Change default location country (Enter to use default): '
            )
    else:
        if user_prompt:
            logger.info('\n')
        logger.info(
            '[collect_transfer_data] Your location could not be determined automatically.'
        )
        if user_prompt:
            location_city = input(
                '[collect_transfer_data] Enter your location city (Enter to skip): '
            )
            location_state = input(
                '[collect_transfer_data] Enter your location state (Enter to skip): '
            )
            location_zip = input(
                '[collect_transfer_data] Enter your location zip (Enter to skip): '
            )
            location_country = input(
                '[collect_transfer_data] Enter your location country (Enter to skip): '
            )
        else:
            location_city = ''
            location_state = ''
            location_zip = ''
            location_country = ''

    # tag = '::'.join([name_first, name_last, location_city, location_state, location_zip, location_country])
    tag_components = [
        name_first,
        name_last,
        location_city,
        location_state,
        location_zip,
        location_country,
    ]
    if offset is not None:
        tag_components += [str(offset)]
    tag_components_clean = [comp.replace(';', '<semi>') for comp in tag_components]
    tag = ','.join(tag_components_clean)
    contributor_rowid = ibs.add_contributors(
        [tag],
        name_first_list=[name_first],
        name_last_list=[name_last],
        loc_city_list=[location_city],
        loc_state_list=[location_state],
        loc_country_list=[location_country],
        loc_zip_list=[location_zip],
    )[0]
    return contributor_rowid


@register_ibs_method
def ensure_contributor_rowids(ibs, user_prompt=False, autolocate=False):
    r"""
    Args:
        ibs (IBEISController):  wbia controller object
        user_prompt (bool):

    Returns:
        list:

    CommandLine:
        python -m wbia.control.manual_meta_funcs --test-ensure_contributor_rowids

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.control.manual_meta_funcs import *  # NOQA
        >>> import wbia
        >>> # build test data
        >>> ibs = wbia.opendb(db='testdb1')
        >>> gid_list = ibs.get_valid_gids()
        >>> ibs.delete_contributors(ibs.get_valid_contributor_rowids())
        >>> contributor_rowid_list1 = ibs.get_image_contributor_rowid(gid_list)
        >>> assert ut.allsame(contributor_rowid_list1)
        >>> ut.assert_eq(contributor_rowid_list1[0], None)
        >>> user_prompt = ut.get_argflag('--user-prompt')
        >>> autolocate = ut.get_argflag('--user-prompt')
        >>> # execute function
        >>> result = ensure_contributor_rowids(ibs, user_prompt, autolocate)
        >>> # verify results
        >>> ibs.print_contributor_table()
        >>> print(result)
        >>> contributor_rowid_list2 = ibs.get_image_contributor_rowid(gid_list)
        >>> assert ut.allsame(contributor_rowid_list2)
        >>> ut.assert_eq(contributor_rowid_list2[0], 1)
    """
    # TODO: Alter this check to support merging databases with more than one contributor, but none assigned to the manual config
    if not ut.QUIET:
        logger.info(
            '[ensure_contributor_rowids] Ensuring all images have contributors for dbname=%r'
            % (ibs.get_dbname())
        )
    contributor_rowid_list = ibs.get_valid_contributor_rowids()
    unassigned_gid_list = ibs.get_all_uncontributed_images()
    if not ut.QUIET:
        logger.info(
            '[ensure_contributor_rowids] %d Contributors exist. %d images are unassigned'
            % (len(contributor_rowid_list), len(unassigned_gid_list))
        )
    if len(unassigned_gid_list) > 0:
        new_contributor_rowid = ibs.add_new_temp_contributor(
            offset=len(contributor_rowid_list),
            user_prompt=user_prompt,
            autolocate=autolocate,
        )
        # SET UNASSIGNED IMAGE CONTRIBUTORS
        ibs.set_image_contributor_rowid(
            unassigned_gid_list, [new_contributor_rowid] * len(unassigned_gid_list)
        )
        # ibs.ensure_imageset_configs_populated()
    # make sure that all images have assigned contributors
    # Get new non-conflicting contributor for unassigned images
    # contributor_rowid_list = list([new_contributor_rowid]) * len(unassigned_gid_list)
    # ibs.set_config_contributor_rowid(unassigned_gid_list, contributor_rowid_list)
    return ibs.get_valid_contributor_rowids()


@register_ibs_method
# @register_api('/api/contributor/gids/uncontributed/', methods=['GET'])
def get_all_uncontributed_images(ibs):
    r"""

    RESTful:
        Method: GET
        URL:    /api/contributor/gids/uncontributed/
    """
    gid_list = ibs.get_valid_gids()
    contributor_rowid_list = ibs.get_image_contributor_rowid(gid_list)
    is_unassigned = [
        contributor_rowid is None for contributor_rowid in contributor_rowid_list
    ]
    unassigned_gid_list = ut.compress(gid_list, is_unassigned)
    return unassigned_gid_list


@register_ibs_method
# @register_api('/api/contributor/configs/uncontributed/', methods=['GET'])
def get_all_uncontributed_configs(ibs):
    r"""

    RESTful:
        Method: GET
        URL:    /api/contributor/configs/uncontributed/
    """
    config_rowid_list = ibs.get_valid_configids()
    contributor_rowid_list = ibs.get_config_contributor_rowid(config_rowid_list)
    isunassigned_list = [
        _contributor_rowid is None for _contributor_rowid in contributor_rowid_list
    ]
    unassigned_config_rowid_list = ut.compress(contributor_rowid_list, isunassigned_list)
    return unassigned_config_rowid_list


@register_ibs_method
# @register_api('/api/config/contributor/unassigned/', methods=['PUT'])
def set_config_contributor_unassigned(ibs, contributor_rowid):
    r"""

    RESTful:
        Method: PUT
        URL:    /api/config/contributor/unassigned/
    """
    # IS THIS NECESSARY?
    unassigned_config_rowid_list = ibs.get_all_uncontributed_configs()
    contributor_rowid_list = [contributor_rowid] * len(unassigned_config_rowid_list)
    ibs.set_config_contributor_rowid(unassigned_config_rowid_list, contributor_rowid_list)


#
# GETTERS::.CONTRIBUTOR_TABLE


@register_ibs_method
@accessor_decors.getter_1to1
# @register_api('/api/contributor/rowid/uuid/', methods=['GET'])
def get_contributor_rowid_from_uuid(ibs, contributor_uuid_list):
    r"""
    Returns:
        contributor_uuid_list (list):  a contributor

    RESTful:
        Method: GET
        URL:    /api/contributor/rowid/uuid/
    """
    # FIXME: MAKE SQL-METHOD FOR NON-ROWID GETTERS
    contributor_rowid_list = ibs.db.get(
        const.CONTRIBUTOR_TABLE,
        ('contributor_rowid',),
        contributor_uuid_list,
        id_colname='contributor_uuid',
    )
    return contributor_rowid_list


@register_ibs_method
@accessor_decors.getter_1to1
@register_api('/api/contributor/rowid/tag/', methods=['GET'])
def get_contributor_rowid_from_tag(ibs, contributor_tag_list):
    r"""
    Returns:
        contributor_tag_list (list):  a contributor

    RESTful:
        Method: GET
        URL:    /api/contributor/rowid/tag/
    """
    # FIXME: MAKE SQL-METHOD FOR NON-ROWID GETTERS
    contributor_rowid_list = ibs.db.get(
        const.CONTRIBUTOR_TABLE,
        ('contributor_rowid',),
        contributor_tag_list,
        id_colname='contributor_tag',
    )
    return contributor_rowid_list


@register_ibs_method
@accessor_decors.getter_1to1
@register_api('/api/contributor/uuid/', methods=['GET'])
def get_contributor_uuid(ibs, contributor_rowid_list):
    r"""
    Returns:
        contributor_uuid_list (list):  a contributor's uuid

    RESTful:
        Method: GET
        URL:    /api/contributor/uuid/
    """
    contributor_uuid_list = ibs.db.get(
        const.CONTRIBUTOR_TABLE, ('contributor_uuid',), contributor_rowid_list
    )
    return contributor_uuid_list


CONTRIBUTOR_LOCATION_CITY = 'contributor_location_city'
CONTRIBUTOR_LOCATION_COUNTRY = 'contributor_location_country'
CONTRIBUTOR_LOCATION_STATE = 'contributor_location_state'
CONTRIBUTOR_LOCATION_ZIP = 'contributor_location_zip'
CONTRIBUTOR_NAME_FIRST = 'contributor_name_first'
CONTRIBUTOR_NAME_LAST = 'contributor_name_last'
CONTRIBUTOR_NOTE = 'contributor_note'
CONTRIBUTOR_ROWID = 'contributor_rowid'
CONTRIBUTOR_TAG = 'contributor_tag'
CONTRIBUTOR_UUID = 'contributor_uuid'
FEATWEIGHT_ROWID = 'featweight_rowid'


def testdata_ibs():
    r""""""
    import wbia

    ibs = wbia.opendb('testdb1')
    qreq_ = None
    return ibs, qreq_


@register_ibs_method
def _get_all_contributor_rowids(ibs):
    r"""
    all_contributor_rowids <- contributor.get_all_rowids()

    Returns:
        list_ (list): unfiltered contributor_rowids

    TemplateInfo:
        Tider_all_rowids
        tbl = contributor

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.control.manual_meta_funcs import *  # NOQA
        >>> ibs, qreq_ = testdata_ibs()
        >>> ibs._get_all_contributor_rowids()
    """
    all_contributor_rowids = ibs.db.get_all_rowids(const.CONTRIBUTOR_TABLE)
    return all_contributor_rowids


@register_ibs_method
@accessor_decors.getter_1to1
@register_api('/api/contributor/tag/', methods=['GET'])
def get_contributor_tag(ibs, contributor_rowid_list, eager=True, nInput=None):
    r"""
    contributor_tag_list <- contributor.contributor_tag[contributor_rowid_list]

    gets data from the "native" column "contributor_tag" in the "contributor" table

    Args:
        contributor_rowid_list (list):

    Returns:
        list: contributor_tag_list -  a contributor's tag

    TemplateInfo:
        Tgetter_table_column
        col = contributor_tag
        tbl = contributor

    CommandLine:
        python -m wbia.templates.template_generator --key contributor  --Tcfg with_api_cache=False with_deleters=False

    RESTful:
        Method: GET
        URL:    /api/contributor/tag/

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.control.manual_meta_funcs import *  # NOQA
        >>> ibs, qreq_ = testdata_ibs()
        >>> contributor_rowid_list = ibs._get_all_contributor_rowids()
        >>> eager = True
        >>> contributor_tag_list = ibs.get_contributor_tag(contributor_rowid_list, eager=eager)
        >>> assert len(contributor_rowid_list) == len(contributor_tag_list)
    """
    id_iter = contributor_rowid_list
    colnames = (CONTRIBUTOR_TAG,)
    contributor_tag_list = ibs.db.get(
        const.CONTRIBUTOR_TABLE,
        colnames,
        id_iter,
        id_colname='rowid',
        eager=eager,
        nInput=nInput,
    )
    return contributor_tag_list


@register_ibs_method
@accessor_decors.getter_1to1
@register_api('/api/contributor/name/first/', methods=['GET'])
def get_contributor_first_name(ibs, contributor_rowid_list):
    r"""
    Returns:
        contributor_name_first_list (list):  a contributor's first name

    RESTful:
        Method: GET
        URL:    /api/contributor/name/first/
    """
    contributor_name_first_list = ibs.db.get(
        const.CONTRIBUTOR_TABLE, ('contributor_name_first',), contributor_rowid_list
    )
    return contributor_name_first_list


@register_ibs_method
@accessor_decors.getter_1to1
@register_api('/api/contributor/name/last/', methods=['GET'])
def get_contributor_last_name(ibs, contributor_rowid_list):
    r"""
    Returns:
        contributor_name_last_list (list):  a contributor's last name

    RESTful:
        Method: GET
        URL:    /api/contributor/name/last/
    """
    contributor_name_last_list = ibs.db.get(
        const.CONTRIBUTOR_TABLE, ('contributor_name_last',), contributor_rowid_list
    )
    return contributor_name_last_list


@register_ibs_method
@accessor_decors.getter_1to1
@register_api('/api/contributor/name/', methods=['GET'])
def get_contributor_name_string(ibs, contributor_rowid_list, include_tag=False):
    r"""
    Returns:
        contributor_name_list (list):  a contributor's full name

    RESTful:
        Method: GET
        URL:    /api/contributor/name/
    """
    first_list = ibs.get_contributor_first_name(contributor_rowid_list)
    last_list = ibs.get_contributor_last_name(contributor_rowid_list)
    if include_tag:
        tag_list = ibs.get_contributor_tag(contributor_rowid_list)
        name_list = zip(first_list, last_list, tag_list)
        contributor_name_list = [
            '%s %s (%s)' % (first, last, tag) for first, last, tag in name_list
        ]
    else:
        name_list = zip(first_list, last_list)
        contributor_name_list = ['%s %s' % (first, last) for first, last in name_list]

    return contributor_name_list


@register_ibs_method
@accessor_decors.getter_1to1
@register_api('/api/contributor/location/city/', methods=['GET'])
def get_contributor_city(ibs, contributor_rowid_list):
    r"""
    Returns:
        contributor_city_list (list):  a contributor's location - city

    RESTful:
        Method: GET
        URL:    /api/contributor/location/city/
    """
    contributor_city_list = ibs.db.get(
        const.CONTRIBUTOR_TABLE, ('contributor_location_city',), contributor_rowid_list
    )
    return contributor_city_list


@register_ibs_method
@accessor_decors.getter_1to1
@register_api('/api/contributor/location/state/', methods=['GET'])
def get_contributor_state(ibs, contributor_rowid_list):
    r"""
    Returns:
        list_ (list):  a contributor's location - state

    RESTful:
        Method: GET
        URL:    /api/contributor/location/state/
    """
    contributor_state_list = ibs.db.get(
        const.CONTRIBUTOR_TABLE, ('contributor_location_state',), contributor_rowid_list
    )
    return contributor_state_list


@register_ibs_method
@accessor_decors.getter_1to1
@register_api('/api/contributor/location/country/', methods=['GET'])
def get_contributor_country(ibs, contributor_rowid_list):
    r"""
    Returns:
        contributor_country_list (list):  a contributor's location - country

    RESTful:
        Method: GET
        URL:    /api/contributor/location/country/
    """
    contributor_country_list = ibs.db.get(
        const.CONTRIBUTOR_TABLE,
        ('contributor_location_country',),
        contributor_rowid_list,
    )
    return contributor_country_list


@register_ibs_method
@accessor_decors.getter_1to1
@register_api('/api/contributor/location/zip/', methods=['GET'])
def get_contributor_zip(ibs, contributor_rowid_list):
    r"""
    Returns:
        contributor_zip_list (list):  a contributor's location - zip

    RESTful:
        Method: GET
        URL:    /api/contributor/location/zip/
    """
    contributor_zip_list = ibs.db.get(
        const.CONTRIBUTOR_TABLE, ('contributor_location_zip',), contributor_rowid_list
    )
    return contributor_zip_list


@register_ibs_method
@accessor_decors.getter_1to1
@register_api('/api/contributor/location/', methods=['GET'])
def get_contributor_location_string(ibs, contributor_rowid_list):
    r"""
    Returns:
        contributor_list (list):  a contributor's location

    RESTful:
        Method: GET
        URL:    /api/contributor/location/
    """
    city_list = ibs.get_contributor_city(contributor_rowid_list)
    state_list = ibs.get_contributor_state(contributor_rowid_list)
    zip_list = ibs.get_contributor_zip(contributor_rowid_list)
    country_list = ibs.get_contributor_country(contributor_rowid_list)
    location_list = zip(city_list, state_list, zip_list, country_list)
    contributor_list = [
        '%s, %s\n%s %s' % (city, state, _zip, country)
        for city, state, _zip, country in location_list
    ]
    return contributor_list


@register_ibs_method
@accessor_decors.getter_1to1
@register_api('/api/contributor/note/', methods=['GET'])
def get_contributor_note(ibs, contributor_rowid_list):
    r"""
    Returns:
        contributor_note_list (list):  a contributor's note

    RESTful:
        Method: GET
        URL:    /api/contributor/note/
    """
    contributor_note_list = ibs.db.get(
        const.CONTRIBUTOR_TABLE, ('contributor_note',), contributor_rowid_list
    )
    return contributor_note_list


@register_ibs_method
@accessor_decors.getter_1to1
# @register_api('/api/contributor/imageset/rowids/', methods=['GET'])
def get_contributor_imgsetids(ibs, config_rowid_list):
    r"""
    Returns:
        imgsetid_list (list):  imgsetids for a contributor

    RESTful:
        Method: GET
        URL:    /api/contributor/imageset/rowids/
    """
    imgsetid_list = ibs.db.get(
        const.IMAGESET_TABLE,
        ('imageset_rowid',),
        config_rowid_list,
        id_colname='config_rowid',
        unpack_scalars=False,
    )
    return imgsetid_list


@register_ibs_method
@accessor_decors.getter_1to1
@register_api('/api/contributor/image/rowid/', methods=['GET'])
def get_contributor_gids(ibs, contributor_rowid_list):
    r"""
    TODO: Template 1_M reverse getter

    Returns:
        gid_list (list):  gids for a contributor

    RESTful:
        Method: GET
        URL:    /api/contributor/gids/
    """
    gid_list = ibs.db.get(
        const.IMAGE_TABLE,
        ('image_rowid',),
        contributor_rowid_list,
        id_colname='contributor_rowid',
        unpack_scalars=False,
    )
    return gid_list


@register_ibs_method
@accessor_decors.getter_1to1
# @register_api('/api/contributor/config/rowid/', methods=['GET'])
def get_config_contributor_rowid(ibs, config_rowid_list):
    r"""
    Returns:
        cfgsuffix_list (list):  contributor's rowid for algorithm configs

    RESTful:
        Method: GET
        URL:    /api/contributor/config/rowid/
    """
    cfgsuffix_list = ibs.db.get(
        const.CONFIG_TABLE, ('contributor_rowid',), config_rowid_list
    )
    return cfgsuffix_list


@register_ibs_method
@accessor_decors.getter_1to1
# @register_api('/api/contributor/config/suffixes/', methods=['GET'])
def get_config_suffixes(ibs, config_rowid_list):
    r"""
    Returns:
        cfgsuffix_list (list):  suffixes for algorithm configs

    RESTful:
        Method: GET
        URL:    /api/contributor/config/suffixes/
    """
    cfgsuffix_list = ibs.db.get(const.CONFIG_TABLE, ('config_suffix',), config_rowid_list)
    return cfgsuffix_list


@register_ibs_method
@accessor_decors.deleter
@register_api('/api/contributor/', methods=['DELETE'])
def delete_contributors(ibs, contributor_rowid_list):
    r"""
    deletes contributors from the database and all information associated

    RESTful:
        Method: DELETE
        URL:    /api/contributor/
    """
    # TODO: FIXME TESTME
    if ut.VERBOSE:
        logger.info('[ibs] deleting %d contributors' % len(contributor_rowid_list))

    # Delete configs (UNSURE IF THIS IS CORRECT)
    # CONTRIBUTORS SHOULD NOT DELETE IMAGES
    # Delete imagesets
    # imgsetid_list = ibs.get_valid_imgsetids()
    # imgsetid_config_list = ibs.get_imageset_configid(imgsetid_list)
    # valid_list = [config in config_rowid_list for config in imgsetid_config_list ]
    # imgsetid_list = ut.compress(imgsetid_list, valid_list)
    # ibs.delete_imagesets(imgsetid_list)
    # Remote image contributors ~~~Delete images~~~~
    gid_list = ut.flatten(ibs.get_contributor_gids(contributor_rowid_list))
    ibs.set_image_contributor_rowid(gid_list, [None] * len(gid_list))
    # Delete contributors
    ibs.db.delete_rowids(const.CONTRIBUTOR_TABLE, contributor_rowid_list)


@register_ibs_method
@accessor_decors.ider
@register_api('/api/contributor/', methods=['GET'])
def get_valid_contributor_rowids(ibs):
    r"""
    Returns:
        list_ (list):  list of all contributor ids

    Returns:
        list: contributor_rowids_list

    CommandLine:
        python -m wbia.control.manual_meta_funcs --test-get_valid_contributor_rowids

    RESTful:
        Method: GET
        URL:    /api/contributor/

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.control.manual_meta_funcs import *  # NOQA
        >>> import wbia
        >>> ibs = wbia.opendb('testdb1')
        >>> contributor_rowids_list = get_valid_contributor_rowids(ibs)
        >>> result = str(contributor_rowids_list)
        >>> print(result)
    """
    contributor_rowids_list = ibs._get_all_contributor_rowids()
    return contributor_rowids_list


@register_ibs_method
@accessor_decors.getter_1to1
# @register_api('/api/metadata/value/', methods=['GET'])
def get_metadata_value(ibs, metadata_key_list, db):
    r"""

    RESTful:
        Method: GET
        URL:    /api/metadata/value/
    """
    params_iter = ((metadata_key,) for metadata_key in metadata_key_list)
    where_clause = 'metadata_key=?'
    # list of relationships for each image
    metadata_value_list = db.get_where(
        const.METADATA_TABLE,
        ('metadata_value',),
        params_iter,
        where_clause,
        unpack_scalars=True,
    )
    return metadata_value_list


@register_ibs_method
@accessor_decors.getter_1to1
# @register_api('/api/metadata/rowid/key/', methods=['GET'])
def get_metadata_rowid_from_metadata_key(ibs, metadata_key_list, db):
    r"""

    RESTful:
        Method: GET
        URL:    /api/metadata/rowid/key/
    """
    db = db[0]  # Unwrap tuple, required by @accessor_decors.getter_1to1 decorator
    params_iter = ((metadata_key,) for metadata_key in metadata_key_list)
    where_clause = 'metadata_key=?'
    # list of relationships for each image
    metadata_rowid_list = db.get_where(
        const.METADATA_TABLE,
        ('metadata_rowid',),
        params_iter,
        where_clause,
        unpack_scalars=True,
    )
    return metadata_rowid_list


@register_api('/api/core/version/', methods=['GET'])
def get_database_version_alias(ibs, db=None):
    r"""
    Alias: `func:get_database_version`

    RESTful:
        Method: GET
        URL:    /api/core/version/
    """
    return ibs.get_database_version(db)


@register_ibs_method
@accessor_decors.ider
@register_api('/api/core/db/version/', methods=['GET'])
def get_database_version(ibs, db=None):
    r"""
    Gets the specified database version from the controller

    RESTful:
        Method: GET
        URL:    /api/core/dbversion/
    """
    if db in [None, 'db', 'core', 'main']:
        db = ibs.db
    elif db in ['staging']:
        db = ibs.staging
    else:
        raise ValueError('Unrecognized database')

    return db.get_db_version()


@register_ibs_method
@accessor_decors.adder
# @register_api('/api/metadata/', methods=['POST'])
def add_metadata(ibs, metadata_key_list, metadata_value_list, db):
    r"""
    Adds metadata

    Returns:
        metadata_rowid_list (list): metadata rowids

    RESTful:
        Method: POST
        URL:    /api/metadata/
    """
    if ut.VERBOSE:
        logger.info('[ibs] adding %d metadata' % len(metadata_key_list))
    # Add imageset text names to database
    colnames = ['metadata_key', 'metadata_value']
    params_iter = zip(metadata_key_list, metadata_value_list)
    get_rowid_from_superkey = functools.partial(
        ibs.get_metadata_rowid_from_metadata_key, db=(db,)
    )
    metadata_rowid_list = db.add_cleanly(
        const.METADATA_TABLE, colnames, params_iter, get_rowid_from_superkey
    )
    return metadata_rowid_list


@register_ibs_method
@profile
def _init_config(ibs):
    r"""
    Loads the database's algorithm configuration

    TODO: per-species config
    """
    #####
    # <GENERAL CONFIG>
    config_fpath = ut.unixjoin(ibs.get_dbdir(), 'general_config.cPkl')
    try:
        general_config = ut.load_cPkl(config_fpath, verbose=ut.VERBOSE)
    except IOError as ex:
        if ut.VERBOSE:
            ut.printex(ex, 'failed to genral load config', iswarning=True)
        general_config = {}
    current_species = general_config.get('current_species', None)
    if ut.VERBOSE and ut.NOT_QUIET:
        logger.info(
            '[_init_config] general_config.current_species = %r' % (current_species,)
        )
    # </GENERAL CONFIG>
    #####
    # species_list = ibs.get_database_species()
    if current_species is None:
        # species_list = ibs.get_database_species()
        # species_list[0] if len(species_list) == 1 else None
        primary_species = ibs.get_primary_database_species()
        current_species = primary_species
    cfgname = 'cfg' if current_species is None else current_species
    if ut.VERBOSE and ut.NOT_QUIET:
        # logger.info('[_init_config] Loading database with species_list = %r ' % (species_list,))
        logger.info('[_init_config] Using cfgname=%r' % (cfgname,))
    # try to be intelligent about the default speceis
    ibs._load_named_config(cfgname)


@register_ibs_method
def _init_burned_in_species(ibs):
    # Add missing "required" species
    species_nice_list = [
        'Giraffe (Masai)',
        'Giraffe (Reticulated)',
        'Other',
        "Zebra (Grevy's)",
        'Zebra (Hybrid)',
        'Zebra (Plains)',
    ]
    species_text_list = [
        'giraffe_masai',
        'giraffe_reticulated',
        'other',
        'zebra_grevys',
        'zebra_hybrid',
        'zebra_plains',
    ]
    species_code_list = [
        'GIRM',
        'GIR',
        'OTHER',
        'GZ',
        'HZ',
        'PZ',
    ]
    ibs.add_species(species_nice_list, species_text_list, species_code_list)
    logger.info('[_init_burned_in_species] Burned in mising species...')


@register_ibs_method
@profile
def _load_named_config(ibs, cfgname=None):
    r""""""
    # TODO: update cfgs between versions
    # Try to load previous config otherwise default
    # use_config_cache = not (ut.is_developer() and not ut.get_argflag(('--nocache-pref',)))
    use_config_cache = not ut.get_argflag(('--nocache-pref',))
    ibs.cfg = Config.load_named_config(cfgname, ibs.get_dbdir(), use_config_cache)
    ibs.reset_table_cache()


@register_ibs_method
def _default_config(ibs, cfgname=None, new=True):
    r"""
    Resets the databases's algorithm configuration

    Args:
        ibs (IBEISController):  wbia controller object
        cfgname (None):

    CommandLine:
        python -m wbia.control.manual_meta_funcs --test-_default_config

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.algo.Config import *  # NOQA
        >>> from wbia.control.manual_meta_funcs import *  # NOQA
        >>> import wbia
        >>> # build test data
        >>> ibs = wbia.opendb('testdb1')
        >>> assert not hasattr(ibs.cfg.query_cfg.flann_cfg, 'badparam')
        >>> ibs.cfg.query_cfg.flann_cfg.badparam = True
        >>> assert ibs.cfg.query_cfg.flann_cfg.cb_index == .4
        >>> ibs.cfg.query_cfg.flann_cfg.cb_index = .5
        >>> assert ibs.cfg.query_cfg.flann_cfg.cb_index == .5
        >>> assert hasattr(ibs.cfg.query_cfg.flann_cfg, 'badparam')
        >>> # bad params is not initially removed but when you specify new it is
        >>> ibs._default_config(new=False)
        >>> assert ibs.cfg.query_cfg.flann_cfg.cb_index == .4
        >>> assert hasattr(ibs.cfg.query_cfg.flann_cfg, 'badparam')
        >>> # bad param should be removed when config is defaulted
        >>> ibs._default_config(new=True)
        >>> assert not hasattr(ibs.cfg.query_cfg.flann_cfg, 'badparam')

    cfg = ibs.cfg
    """
    # species_list = ibs.get_database_species()
    # if len(species_list) == 1:
    #    # try to be intelligent about the default speceis
    #    cfgname = species_list[0]
    ibs.cfg = Config._default_config(ibs.cfg, cfgname, new=new)
    ibs.reset_table_cache()


@register_ibs_method
@register_api('/api/query/cfg/', methods=['PUT'])
def update_query_cfg(ibs, **kwargs):
    r"""
    Updates query config only. Configs needs a restructure very badly
    DEPRICATE

    RESTful:
        Method: PUT
        URL:    /api/query/cfg/
    """
    Config.update_query_config(ibs.cfg, **kwargs)
    ibs.reset_table_cache()



# -*- coding: utf-8 -*-
"""
Interface to Faster R-CNN object proposals.
"""
import logging
import utool as ut
import vtool as vt
from six.moves import zip, range
from os.path import abspath, dirname, expanduser, join, exists  # NOQA
import numpy as np
import sys
import cv2

(print, rrr, profile) = ut.inject2(__name__, '[faster r-cnn]')
logger = logging.getLogger('wbia')

# SCRIPT_PATH = abspath(dirname(__file__))
SCRIPT_PATH = abspath(expanduser(join('~', 'code', 'py-faster-rcnn')))

if not ut.get_argflag('--no-faster-rcnn'):
    try:
        assert exists(SCRIPT_PATH)

        def add_path(path):
            # if path not in sys.path:
            sys.path.insert(0, path)

        # Add pycaffe to PYTHONPATH
        pycaffe_path = join(SCRIPT_PATH, 'caffe-fast-rcnn', 'python')
        add_path(pycaffe_path)

        # Add caffe lib path to PYTHONPATH
        lib_path = join(SCRIPT_PATH, 'lib')
        add_path(lib_path)

        import caffe

        ut.reload_module(caffe)
        from fast_rcnn.config import cfg
        from fast_rcnn.test import im_detect

        # from fast_rcnn.nms_wrapper import nms
    except AssertionError:
        logger.info(
            'WARNING Failed to find py-faster-rcnn. ' 'Faster R-CNN is unavailable'
        )
        # if ut.SUPER_STRICT:
        #     raise
    except ImportError:
        logger.info('WARNING Failed to import fast_rcnn. ' 'Faster R-CNN is unavailable')
        # if ut.SUPER_STRICT:
        #     raise


VERBOSE_SS = ut.get_argflag('--verbdss') or ut.VERBOSE


CONFIG_URL_DICT = {
    # 'pretrained-fast-vgg-pascal' : 'https://wildbookiarepository.azureedge.net/models/pretrained.fastrcnn.vgg16.pascal.prototxt',  # Trained on PASCAL VOC 2007
    'pretrained-vgg-pascal': 'https://wildbookiarepository.azureedge.net/models/pretrained.fasterrcnn.vgg16.pascal.prototxt',  # Trained on PASCAL VOC 2007
    'pretrained-zf-pascal': 'https://wildbookiarepository.azureedge.net/models/pretrained.fasterrcnn.zf.pascal.prototxt',  # Trained on PASCAL VOC 2007
    'pretrained-vgg-ilsvrc': 'https://wildbookiarepository.azureedge.net/models/pretrained.fasterrcnn.vgg16.ilsvrc.prototxt',  # Trained on ILSVRC 2014
    'pretrained-zf-ilsvrc': 'https://wildbookiarepository.azureedge.net/models/pretrained.fasterrcnn.zf.ilsvrc.prototxt',  # Trained on ILSVRC 2014
    'default': 'https://wildbookiarepository.azureedge.net/models/pretrained.fasterrcnn.vgg16.pascal.prototxt',  # Trained on PASCAL VOC 2007
    None: 'https://wildbookiarepository.azureedge.net/models/pretrained.fasterrcnn.vgg16.pascal.prototxt',  # Trained on PASCAL VOC 2007
}


def _parse_weight_from_cfg(url):
    return url.replace('.prototxt', '.caffemodel')


def _parse_classes_from_cfg(url):
    return url.replace('.prototxt', '.classes')


def _parse_class_list(classes_filepath):
    # Load classes from file into the class list
    assert exists(classes_filepath)
    class_list = []
    with open(classes_filepath) as classes:
        for line in classes.readlines():
            line = line.strip()
            if len(line) > 0:
                class_list.append(line)
    return class_list


def detect_gid_list(ibs, gid_list, downsample=True, verbose=VERBOSE_SS, **kwargs):
    """
    Args:
        gid_list (list of int): the list of IBEIS image_rowids that need detection
        downsample (bool, optional): a flag to indicate if the original image
            sizes should be used; defaults to True

            True:  ibs.get_image_detectpaths() is used
            False: ibs.get_image_paths() is used

    Kwargs (optional): refer to the Faster R-CNN documentation for configuration settings

    Args:
        ibs (wbia.IBEISController):  image analysis api
        gid_list (list of int): the list of IBEIS image_rowids that need detection
        downsample (bool, optional): a flag to indicate if the original image
                sizes should be used; defaults to True

    Kwargs:
        detector, config_filepath, weights_filepath, verbose

    Yields:
        tuple: (gid, gpath, result_list)

    CommandLine:
        python -m wbia.algo.detect.fasterrcnn detect_gid_list --show

    Example:
        >>> # DISABLE_DOCTEST
        >>> from wbia.algo.detect.fasterrcnn import *  # NOQA
        >>> from wbia.core_images import LocalizerConfig
        >>> import wbia
        >>> ibs = wbia.opendb('testdb1')
        >>> gid_list = ibs.get_valid_gids()
        >>> config = {'verbose': True}
        >>> downsample = False
        >>> results_list = detect_gid_list(ibs, gid_list, downsample, **config)
        >>> results_list = list(results_list)
        >>> print('result lens = %r' % (map(len, list(results_list))))
        >>> print('result[0] = %r' % (len(list(results_list[0][2]))))
        >>> config = {'verbose': True}
        >>> downsample = False
        >>> results_list = detect_gid_list(ibs, gid_list, downsample, **config)
        >>> results_list = list(results_list)
        >>> print('result lens = %r' % (map(len, list(results_list))))
        >>> print('result[0] = %r' % (len(list(results_list[0][2]))))
        >>> ut.quit_if_noshow()
        >>> import wbia.plottool as pt
        >>> ut.show_if_requested()

    Yields:
        results (list of dict)
    """
    # Get new gpaths if downsampling
    if downsample:
        gpath_list = ibs.get_image_detectpaths(gid_list)
        neww_list = [vt.open_image_size(gpath)[0] for gpath in gpath_list]
        oldw_list = [oldw for (oldw, oldh) in ibs.get_image_sizes(gid_list)]
        downsample_list = [oldw / neww for oldw, neww in zip(oldw_list, neww_list)]
        orient_list = [1] * len(gid_list)
    else:
        gpath_list = ibs.get_image_paths(gid_list)
        downsample_list = [None] * len(gpath_list)
        orient_list = ibs.get_image_orientation(gid_list)
    # Run detection
    results_iter = detect(gpath_list, verbose=verbose, **kwargs)
    # Upscale the results
    _iter = zip(downsample_list, gid_list, orient_list, results_iter)
    for downsample, gid, orient, (gpath, result_list) in _iter:
        # Upscale the results back up to the original image size
        for result in result_list:
            if downsample is not None and downsample != 1.0:
                for key in ['xtl', 'ytl', 'width', 'height']:
                    result[key] = int(result[key] * downsample)
            bbox = (
                result['xtl'],
                result['ytl'],
                result['width'],
                result['height'],
            )
            bbox_list = [bbox]
            bbox = bbox_list[0]
            result['xtl'], result['ytl'], result['width'], result['height'] = bbox
        yield (gid, gpath, result_list)


def detect(
    gpath_list,
    config_filepath,
    weight_filepath,
    class_filepath,
    sensitivity,
    verbose=VERBOSE_SS,
    use_gpu=True,
    use_gpu_id=0,
    **kwargs,
):
    """
    Args:
        gpath_list (list of str): the list of image paths that need proposal candidates

    Kwargs (optional): refer to the Faster R-CNN documentation for configuration settings

    Returns:
        iter
    """
    cfg.TEST.HAS_RPN = True  # Use RPN for proposals

    # Get correct config if specified with shorthand
    config_url = None
    if config_filepath in CONFIG_URL_DICT:
        config_url = CONFIG_URL_DICT[config_filepath]
        config_filepath = ut.grab_file_url(config_url, appname='wbia', check_hash=True)

    # Get correct weights if specified with shorthand
    if weight_filepath in CONFIG_URL_DICT:
        if weight_filepath is None and config_url is not None:
            config_url_ = config_url
        else:
            config_url_ = CONFIG_URL_DICT[weight_filepath]
        weight_url = _parse_weight_from_cfg(config_url_)
        weight_filepath = ut.grab_file_url(weight_url, appname='wbia', check_hash=True)

    if class_filepath is None:
        class_url = _parse_classes_from_cfg(config_url)
        class_filepath = ut.grab_file_url(
            class_url, appname='wbia', check_hash=True, verbose=verbose
        )
    class_list = _parse_class_list(class_filepath)

    # Need to convert unicode strings to Python strings to support Boost Python
    # call signatures in caffe
    prototxt_filepath = str(config_filepath)  # alias to Caffe nomenclature
    caffemodel_filepath = str(weight_filepath)  # alias to Caffe nomenclature

    assert exists(prototxt_filepath), 'Specified prototxt file not found'
    assert exists(caffemodel_filepath), 'Specified caffemodel file not found'

    if use_gpu:
        caffe.set_mode_gpu()
        caffe.set_device(use_gpu_id)
        cfg.GPU_ID = use_gpu_id
    else:
        caffe.set_mode_cpu()

    net = caffe.Net(prototxt_filepath, caffemodel_filepath, caffe.TEST)

    # Warm-up network on a dummy image
    im = 128 * np.ones((300, 500, 3), dtype=np.uint8)
    for i in range(2):
        _, _ = im_detect(net, im)

    results_list_ = []
    for gpath in gpath_list:
        image = cv2.imread(gpath)
        score_list, bbox_list = im_detect(net, image)

        # Compile results
        result_list_ = []
        for class_index, class_name in enumerate(class_list[1:]):
            class_index += 1  # because we skipped background
            class_boxes = bbox_list[:, 4 * class_index : 4 * (class_index + 1)]
            class_scores = score_list[:, class_index]
            dets_list = np.hstack((class_boxes, class_scores[:, np.newaxis]))
            dets_list = dets_list.astype(np.float32)
            # # Perform NMS
            # keep_list = nms(dets_list, nms_sensitivity)
            # dets_list = dets_list[keep_list, :]
            # Perform sensitivity check
            keep_list = np.where(dets_list[:, -1] >= sensitivity)[0]
            dets_list = dets_list[keep_list, :]
            for (xtl, ytl, xbr, ybr, conf) in dets_list:
                xtl = int(np.around(xtl))
                ytl = int(np.around(ytl))
                xbr = int(np.around(xbr))
                ybr = int(np.around(ybr))
                confidence = float(conf)
                result_dict = {
                    'xtl': xtl,
                    'ytl': ytl,
                    'width': xbr - xtl,
                    'height': ybr - ytl,
                    'class': class_name,
                    'confidence': confidence,
                }
                result_list_.append(result_dict)
        results_list_.append(result_list_)

    results_list = zip(gpath_list, results_list_)
    return results_list



# -*- coding: utf-8 -*-
"""Dependencies: flask, tornado."""
import logging
from wbia.control import controller_inject
from flask_swagger import swagger
import wbia.constants as const
from flask import current_app
from flask import jsonify
from flask import url_for
import utool as ut
import traceback
import uuid

(print, rrr, profile) = ut.inject2(__name__)
logger = logging.getLogger('wbia')

CLASS_INJECT_KEY, register_ibs_method = controller_inject.make_ibs_register_decorator(
    __name__
)

PREFIX = controller_inject.MICROSOFT_API_PREFIX
register_api = controller_inject.get_wbia_flask_api(__name__)
register_route = controller_inject.get_wbia_flask_route(__name__)


def _prefix(route=''):
    rule = '/%s/%s/' % (PREFIX, route)
    while '//' in rule:
        rule = rule.replace('//', '/')
    return rule


def _image(ibs, gid):
    return {
        'uuid': str(ibs.get_image_uuids(gid)),
    }


def _annotation(ibs, aid):
    return {
        'uuid': str(ibs.get_annot_uuids(aid)),
    }


def _name(ibs, nid):
    return {
        'uuid': str(ibs.get_name_uuids(nid)),
    }


# DEPRICATE
#     def _detection(ibs, aid):
#         gid = ibs.get_annot_gids(aid)
#         bbox = ibs.get_annot_bboxes(aid)
#         return {
#             '_image'      : _image(ibs, gid),
#             '_annotation' : _annotation(ibs, aid),
#             'xtl'         : bbox[0],
#             'ytl'         : bbox[1],
#             'width'       : bbox[2],
#             'height'      : bbox[3],
#             'theta'       : ibs.get_annot_thetas(aid),
#             'label'       : ibs.get_annot_species_texts(aid),
#             'score'       : ibs.get_annot_detect_confidence(aid),
#         }


def _detection(ibs, gid, result):
    bbox, theta, conf, label = result
    detection = {
        'image': _image(ibs, gid),
        'bbox': bbox,
        'xtl': bbox[0],
        'ytl': bbox[1],
        'xbr': bbox[0] + bbox[2],
        'ybr': bbox[1] + bbox[3],
        'width': bbox[2],
        'height': bbox[3],
        'theta': theta,
        'species': label,
        'viewpoint': None,
        'score': conf,
    }
    return detection


def _task(ibs, taskid):
    return {
        'uuid': taskid,
    }


def _ensure_general(ibs, models, tag, rowid_from_uuid_func, unpack=True, *args, **kwargs):

    if isinstance(models, dict):
        models = [models]

    length = len(models)
    single = length == 1
    if length == 0:
        return []

    rowid_list = []
    for index, model in enumerate(models):
        try:
            if single:
                parameter = 'models'
            else:
                parameter = 'models:%d' % (index,)

            assert 'uuid' in model, '%s Model provided is invalid, missing UUID key' % (
                tag,
            )

            uuid_ = uuid.UUID(model['uuid'])
            assert uuid_ is not None, "%s Model's UUID is invalid" % (tag,)

            rowid = rowid_from_uuid_func(uuid_)
            assert rowid is not None, '%s Model is unrecognized, please upload' % (tag,)
        except AssertionError as ex:
            raise controller_inject.WebInvalidInput(str(ex), parameter)
        rowid_list.append(rowid)

    if single and unpack:
        return rowid_list[0]
    else:
        return rowid_list


def _ensure_images(ibs, images, *args, **kwargs):
    return _ensure_general(
        ibs, images, 'Image', ibs.get_image_gids_from_uuid, *args, **kwargs
    )


def _ensure_annotations(ibs, annotations, *args, **kwargs):
    return _ensure_general(
        ibs, annotations, 'Annotation', ibs.get_annot_aids_from_uuid, *args, **kwargs
    )


def _ensure_names(ibs, names, *args, **kwargs):
    return _ensure_general(
        ibs, names, 'Name', ibs.get_name_rowids_from_uuid, *args, **kwargs
    )


@register_route(_prefix('swagger'), methods=['GET'])
def microsoft_core_specification_swagger(*args, **kwargs):
    r"""
    Returns the API specification in the Swagger 2.0 (OpenAPI) JSON format.

    The Swagger API specification (https://github.com/OAI/OpenAPI-Specification/blob/master/versions/2.0.md) provides a standardized method to export REST API documentation and examples.  Our documentation is built on-demand with the help of the Python package flask-swagger (https://github.com/gangverk/flask-swagger).

    The API specification includes GET, POST, PUT, and DELETE methods and Model definitions.
    ---
    definitions:
    - schema:
        id: Image
        description: An Image is a semantic construct that represents an uploaded image.  Images can be uploaded for later processing or be used immediately for detection.  Object detection will create Annotation models, which have a required Image parent.  An Image can have multiple detections (i.e. Annotation models).
        required:
          - uuid
        properties:
          uuid:
            description: a deterministically-derived UUID based on the image pixels, which can be used to identify duplicate Images.
            type: string
            format: uuid
    - schema:
        id: Annotation
        description: An Annotation is a semantic construct that represents a *committed* detection, with a bounding box and species classification assigned as stored attributes.  An Annotation is required to have a parent Image.  All bounding box coordinates are relative to the size of the parent Image that was uploaded.
        required:
          - uuid
          - bbox
        properties:
          uuid:
            description: a deterministically-derived UUID based on the parent image's UUID and the bounding box coordinate (xtl, ytl, width, height) and orientation (theta), which can be used to identify duplicate Annotations.
            type: string
            format: uuid
          bbox:
            description: a 4-tuple of coordinates that defines a rectangle in the format (x-axis top left corner, y-axis top left corner, width, height) in pixels.  These values are expected to be bounded by the size of the parent image.
            type: array
          theta:
            description: a rotation around the center of the annotation, in radians
            type: number
          species:
            description: a user-defined string to specify the species of the annotation (e.g. 'zebra' or 'massai_giraffe').  This value is used to filter matches and run-time models for ID.
            type: string
          viewpoint:
            description: a user-defined string to specify the viewpoint of the annotation (e.g. 'right' or 'front_left').  This value is used to filter matches and run-time models for ID.
            type: string
          name:
            description: the name of the individual
            format: uuid
            type: string
    - schema:
        id: Detection
        description: A Detection is a semantic constrict that represents an *un-committed* detection.  A Detection can be committed to an Annotation to be stored permanently on the parent Image.
        required:
          - _image
          - score
          - bbox
          - xtl
          - ytl
          - xbr
          - ybr
          - width
          - height
          - theta
          - label
        properties:
          image:
            description: The Image that this Detection was found in
            $ref: "#/definitions/Image"
          score:
            description: The detection's classification score
            type: integer
            format: int32
          bbox:
            description: The bounding box for this annotation, represented in the format (xtl, ytl, width, height)
            type: array
            items:
              type: number
              format: float
          xtl:
            description: The pixel coordinate for the top-left corner along the x-axis (xtl = x-axis top left) for the bounding box
            type: integer
            format: int32
          ytl:
            description: The pixel coordinate for the top-left corner along the y-axis (ytl = y-axis top left) for the bounding box
            type: integer
            format: int32
          xbr:
            description: The pixel coordinate for the bottom-right corner along the x-axis (ytl = x-axis bottom right) for the bounding box
            type: integer
            format: int32
          ybr:
            description: The pixel coordinate for the bottom-right corner along the y-axis (ytl = y-axis bottom right) for the bounding box
            type: integer
            format: int32
          width:
            description: The pixel width for the bounding box
            type: integer
            format: int32
          height:
            description: The pixel height for the bounding box
            type: integer
            format: int32
          theta:
            description: The rotation of the bounding box around its center, represented in radians
            type: number
            format: float
          species:
            description: The semantic species classification (class label) of the bounding box
            type: string
          viewpoint:
            description: The semantic viewpoint classification (class label) of the bounding box
            type: string
    - schema:
        id: Name
        description: A Name is the identification label for a group of Annotations
        required:
          - uuid
          - alias
        properties:
          uuid:
            description: a deterministically-derived UUID based on the image pixels, which can be used to identify duplicate Images.
            type: string
            format: uuid
          alias:
            description: a string alias for this individual, helpful for user-facing interfaces
            type: string
    - schema:
        id: Task
        description: A Task is a semantic construct that represents a background task (i.e. detection) in an asynchronous call.  A Task has an optional callback on completion or the status (and result) can be checked via the API
        required:
          - uuid
        properties:
          uuid:
            description: a random UUID to identify a given asynchronous call, used to check status and results of a background task
            type: string
            format: uuid
    produces:
    - application/json
    responses:
        200:
          description: Returns the Swagger 2.0 JSON format
    """
    try:
        swag = swagger(current_app)
    except Exception:
        logger.info(str(traceback.format_exc()))
        # ut.embed()

    swag['info']['title'] = 'Wild Me - IA (Image Analysis)'
    swag['info'][
        'description'
    ] = 'Documentation for all classification, detection, and identification calls provided by Wild Me for the AI for Earth (AI4E) collaboration'
    swag['info']['version'] = 'v0.1'
    swag['info']['contact'] = {
        'name': 'Wild Me Developers (AI4E)',
        'url': 'http://wildme.org',
        'email': 'dev@wildme.org',
    }
    swag['info']['license'] = {
        'name': 'Apache 2.0',
        'url': 'http://www.apache.org/licenses/LICENSE-2.0.html',
    }
    swag['host'] = 'demo.wildbook.org:5010'
    swag['schemes'] = [
        'http',
    ]
    # swag['basePath'] = PREFIX

    # "securityDefinitions": {
    #   "apiKeyHeader": {
    #     "type": "apiKey",
    #     "name": "Ocp-Apim-Subscription-Key",
    #     "in": "header"
    #   },
    #   "apiKeyQuery": {
    #     "type": "apiKey",
    #     "name": "subscription-key",
    #     "in": "query"
    #   }
    # },
    # "security": [
    #   {
    #     "apiKeyHeader": []
    #   },
    #   {
    #     "apiKeyQuery": []
    #   }
    # ],

    response = jsonify(swag)
    return response


@register_api(_prefix('status'), methods=['GET'], __api_plural_check__=False)
def microsoft_core_status(ibs, *args, **kwargs):
    r"""
    Returns the health status of the API back-end; optionally can be used as a service heatbeat.
    ---
    produces:
    - application/json
    responses:
      200:
        description: Returns the status of the server
        schema:
          type: object
          properties:
            status:
              type: string
              enum:
              - healthy
              - warning
              - critical
        examples:
          application/json:
            status: healthy
    """
    ibs.heartbeat()
    status = 'healthy'
    return {'status': status}


@register_api(_prefix('image'), methods=['POST'])
def microsoft_image_upload(ibs, *args, **kwargs):
    r"""
    Upload an Image to the system and return the UUID
    ---
    parameters:
    - name: image
      in: formData
      description: The image to upload.
      required: true
      type: file
      enum:
      - image/png
      - image/jpg
      - image/tiff
    consumes:
    - multipart/form-data
    produces:
    - application/json
    responses:
      200:
        description: Returns an Image model with a UUID
        schema:
          $ref: "#/definitions/Image"
      400:
        description: Invalid input parameter
      415:
        description: Unsupported media type in the request body. Currently only image/png, image/jpeg, image/tiff are supported.
    """
    from wbia.web.apis import image_upload

    try:
        gid = image_upload(cleanup=True, **kwargs)
        assert gid is not None
    except controller_inject.WebException:
        raise
    except Exception:
        raise controller_inject.WebInvalidInput(
            'Uploaded image is corrupted or is an unsupported file format (supported: image/png, image/jpeg, image/tiff)',
            'image',
            image=True,
        )
    return _image(ibs, gid)


@register_api(_prefix('annotation'), methods=['POST'])
def microsoft_annotation_add(
    ibs,
    image,
    bbox,
    theta=None,
    species=None,
    viewpoint=None,
    name=None,
    *args,
    **kwargs,
):
    r"""
    Add an Annotation to the system and return the UUID
    ---
    parameters:
    - name: image
      in: body
      description: A Image model for which the annotation will be added as a child
      required: true
      schema:
        $ref: "#/definitions/Image"
    - name: bbox
      in: formData
      description: a 4-tuple of coordinates that defines a rectangle in the format (x-axis top left corner, y-axis top left corner, width, height) in pixels.  These values are expected to be bounded by the size of the parent image.
      required: true
      type: array
      items:
        type: number
        format: float
    - name: theta
      in: formData
      description: a rotation around the center of the annotation, in radians.  Default is 0.0
      required: false
      type: number
    - name: species
      in: formData
      description: a user-defined string to specify the species of the annotation (e.g. 'zebra' or 'massai_giraffe').  This value is used to filter matches and run-time models for ID.  Default is None.
      required: false
      type: string
    - name: viewpoint
      in: formData
      description: a user-defined string to specify the viewpoint of the annotation (e.g. 'right' or 'front_left').  This value is used to filter matches and run-time models for ID. Default is None.
      required: false
      type: string
    - name: name
      in: body
      description: a Name model to assign to the created Annotation
      required: false
      schema:
        $ref: "#/definitions/Name"
    consumes:
    - multipart/form-data
    produces:
    - application/json
    responses:
      200:
        description: Returns an Annotations model with a UUID
        schema:
          $ref: "#/definitions/Annotation"
      400:
        description: Invalid input parameter
    """
    try:
        parameter = 'bbox'
        assert isinstance(bbox, (tuple, list)), 'Bounding box must be a list'
        assert len(bbox) == 4, 'Bounding Box must have 4 yntegers'
        assert isinstance(bbox[0], int), 'Bounding box xtl (index 0) must be an integer'
        assert isinstance(bbox[1], int), 'Bounding box ytl (index 1) must be an integer'
        assert isinstance(bbox[2], int), 'Bounding box width (index 2) must be an integer'
        assert isinstance(
            bbox[3], int
        ), 'Bounding box height (index 3) must be an integer'

        if theta is not None:
            assert isinstance(theta, float), 'Theta must be a float'

        if species is not None:
            assert isinstance(species, str), 'Species must be a string'
            assert len(species) > 0, 'Species cannot be empty'

        if viewpoint is not None:
            assert isinstance(viewpoint, str), 'Viewpoint must be a string'
            assert len(viewpoint) > 0, 'Viewpoint cannot be empty'
            assert (
                viewpoint in const.YAWALIAS
            ), 'Invalid viewpoint provided.  Must be one of: %s' % (
                list(const.YAWALIAS.keys()),
            )

        if name is not None:
            nid = _ensure_names(ibs, name)
        else:
            nid = None
    except AssertionError as ex:
        raise controller_inject.WebInvalidInput(str(ex), parameter)

    gid = _ensure_images(ibs, image)

    gid_list = [gid]
    bbox_list = [bbox]
    theta_list = None if theta is None else [theta]
    species_list = None if species is None else [species]
    viewpoint_list = None if viewpoint is None else [viewpoint]
    nid_list = None if nid is None else [nid]

    aid_list = ibs.add_annots(
        gid_list,
        bbox_list=bbox_list,
        theta_list=theta_list,
        species_list=species_list,
        viewpoint_list=viewpoint_list,
        nid_list=nid_list,
    )
    assert len(aid_list) == 1
    aid = aid_list[0]

    response = _annotation(ibs, aid)
    logger.info(response)
    return response


@register_api(_prefix('name'), methods=['POST'])
def microsoft_name_add(ibs, alias, *args, **kwargs):
    r"""
    Add a Name to the system and return the UUID
    ---
    parameters:
    - name: alias
      in: formData
      description: a user-defined string to specify an alias for this name, useful for forward facing interfaces
      required: false
      type: string
    consumes:
    - multipart/form-data
    produces:
    - application/json
    responses:
      200:
        description: Returns an Name model with a UUID
        schema:
          $ref: "#/definitions/Name"
      400:
        description: Invalid input parameter
    """
    try:
        parameter = 'alias'
        assert isinstance(alias, str), 'Alias must be a string'
        assert len(alias) > 0, 'Alias cannot be empty'
    except AssertionError as ex:
        raise controller_inject.WebInvalidInput(str(ex), parameter)

    nid = ibs.add_names(alias)
    return _name(ibs, nid)


@register_api(_prefix('image/annotations'), methods=['GET'], __api_plural_check__=False)
def microsoft_image_annotations(ibs, image, *args, **kwargs):
    r"""
    Get the Annotation models for a given Image model
    ---
    parameters:
    - name: image
      in: body
      description: A Image model to use for querying associated Annotation UUIDs
      required: true
      schema:
        $ref: "#/definitions/Image"
    produces:
    - application/json
    responses:
      200:
        description: Returns a (possibly empty) list of Annotations models
        schema:
          type: array
          items:
            $ref: "#/definitions/Annotation"
      400:
        description: Invalid input parameter
    """
    gid = _ensure_images(ibs, image)
    aid_list = ibs.get_image_aids(gid)

    annotation_list = [_annotation(ibs, aid) for aid in aid_list]
    response = {
        'annotations': annotation_list,
    }
    return response


@register_api(_prefix('name/annotations'), methods=['GET'], __api_plural_check__=False)
def microsoft_name_annotations(ibs, name, *args, **kwargs):
    r"""
    Get the Annotation models for a given Name model
    ---
    parameters:
    - name: name
      in: body
      description: A Name model to use for querying associated Annotation UUIDs
      required: true
      schema:
        $ref: "#/definitions/Name"
    produces:
    - application/json
    responses:
      200:
        description: Returns a (possibly empty) list of Annotations models
        schema:
          type: array
          items:
            $ref: "#/definitions/Annotation"
      400:
        description: Invalid input parameter
    """
    nid = _ensure_names(ibs, name)
    aid_list = ibs.get_name_aids(nid)

    annotation_list = [_annotation(ibs, aid) for aid in aid_list]
    response = {
        'annotations': annotation_list,
    }
    return response


@register_api(_prefix('annotation'), methods=['GET'])
def microsoft_annotation_metadata(ibs, annotation, *args, **kwargs):
    r"""
    Get the Annotation's metadata (bbox, theta, species, viewpoint).

    We provide no setters for Annotations.  If you wish to modify the metadata on an Annotation model, please delete it and add it again.  Modifying the metadata on an annotation may alter its derived UUID
    ---
    parameters:
    - name: annotation
      in: body
      description: A Annotation model
      required: true
      schema:
        $ref: "#/definitions/Image"
    produces:
    - application/json
    responses:
      200:
        description: Returns a JSON object with the metadata for the annotation
        schema:
          type: object
      400:
        description: Invalid input parameter
    """
    aid = _ensure_annotations(ibs, annotation)
    gid = ibs.get_annot_gids(aid)
    image = _image(ibs, gid)
    nid = ibs.get_annot_nids(aid)
    if nid is None or nid <= 0:
        name = None
    else:
        name = _name(ibs, nid)

    metadata = {
        'image': image,
        'name': name,
        'bbox': ibs.get_annot_bboxes(aid),
        'theta': ibs.get_annot_thetas(aid),
        'species': ibs.get_annot_species_texts(aid),
        'viewpoint': ibs.get_annot_viewpoints(aid),
    }
    return metadata


@register_api(_prefix('detect'), methods=['GET'])
def microsoft_detect_model(ibs, *args, **kwargs):
    r"""
    Returns the available models and their supported species for detection.  These models are pre-trained and are downloaded as needed on first start-up.

    The current model names that are supported for demoing and their species:
    - hammerhead -> shark_hammerhead
    - jaguar     -> jaguar
    - lynx       -> lynx
    - manta      -> manta_ray_giant
    - seaturtle  -> fish, ignore, person, turtle_green, turtle_green+head, turtle_hawksbill, turtle_hawksbill+head
    - ggr2       -> giraffe, zebra
    ---
    produces:
    - application/json
    responses:
      200:
        description: Returns an object of models (keys) and their supported species (values) listed as an array of strings.
        schema:
          type: object
    """
    hidden_models = [
        'hendrik_elephant',
        'hendrik_elephant_ears',
        'hendrik_elephant_ears_left',
        'hendrik_dorsal',
        'candidacy',
        None,
    ]
    return ibs.models_cnn_lightnet(hidden_models=hidden_models)


def microsoft_detect_input_validation(model, score_threshold, use_nms, nms_threshold):
    from wbia.algo.detect.lightnet import CONFIG_URL_DICT

    try:
        parameter = 'model'
        assert model in CONFIG_URL_DICT, 'Specified model is not supported'

        parameter = 'score_threshold'
        assert isinstance(score_threshold, float), 'Score threshold must be a float'
        assert (
            0.0 <= score_threshold and score_threshold <= 1.0
        ), 'Score threshold is invalid, must be in range [0.0, 1.0]'

        parameter = 'use_nms'
        assert isinstance(use_nms, bool), 'NMS flag must be a boolean'

        parameter = 'nms_threshold'
        assert isinstance(nms_threshold, float), 'NMS threshold must be a float'
        assert (
            0.0 <= nms_threshold and nms_threshold <= 1.0
        ), 'NMS threshold is invalid, must be in range [0.0, 1.0]'
    except AssertionError as ex:
        raise controller_inject.WebInvalidInput(str(ex), parameter)


@register_ibs_method
def microsoft_detect(
    ibs,
    images,
    model,
    score_threshold=0.0,
    use_nms=True,
    nms_threshold=0.4,
    __jobid__=None,
    *args,
    **kwargs,
):
    depc = ibs.depc_image

    gid_list = _ensure_images(ibs, images, unpack=False)

    try:
        config = {
            'algo': 'lightnet',
            'config_filepath': model,
            'weight_filepath': model,
            'sensitivity': score_threshold,
            'nms': use_nms,
            'nms_thresh': nms_threshold,
        }
        results_list = depc.get_property('localizations', gid_list, None, config=config)

        # DEPRICATE
        #     # Do not commit results, return them outright
        #     aids_list = ibs.commit_localization_results(gid_list, results_list)
    except Exception:
        logger.info(str(traceback.format_exc()))
        raise controller_inject.WebException(
            'Detection process failed for an unknown reason'
        )

    # DEPRICATE
    #     # Do not commit detections
    #     detections_list = {
    #         'detections_list': [
    #             {
    #                 'detections' : [
    #                     _detection(ibs, aid)
    #                     for aid in aid_list
    #                 ],
    #             }
    #             for aid_list in aids_list
    #         ]
    #     }

    zipped = zip(gid_list, results_list)
    detections_list = {
        'detections_list': [
            {
                'detections': [
                    _detection(ibs, gid, result) for result in zip(*result_list[1:])
                ],
            }
            for gid, result_list in zipped
        ]
    }

    return detections_list


@register_api(_prefix('detect'), methods=['POST'])
def microsoft_detect_upload(
    ibs, model, score_threshold=0.0, use_nms=True, nms_threshold=0.4, *args, **kwargs
):
    r"""
    Returns the detection results for an uploaded image and a provided model configuration.

    The uploaded image will be used to perform object detection for a given model.  A detection will return at most 845 Detections (13 x 13 grid of 5 bounding boxes based on the YOLO v2 by Redmon et al.).
    ---
    parameters:
    - name: image
      in: formData
      description: The image for which detection will be processed.
      required: true
      type: file
      enum:
      - image/png
      - image/jpg
      - image/tiff
    - name: model
      in: formData
      description: The model name to use with detection.  Available models can be queried using the API $PREFIX/detect/model/
      required: true
      type: string
    - name: score_threshold
      in: formData
      description: A score threshold to filter the detection results.  Default 0.0.  Must be in range [0, 1], inclusive
      required: false
      type: number
      format: float
    - name: use_nms
      in: formData
      description: Flag to turn on/off Non-Maximum Suppression (NMS).  Default True.  Must be in range [0, 1], inclusive
      required: false
      type: boolean
    - name: nms_threshold
      in: formData
      description: A Intersection-over-Union (IoU) threshold to suppress detection results with NMS.  Default 0.0.  Must be in range [0, 1], inclusive
      required: false
      type: number
      format: float
    consumes:
    - multipart/form-data
    responses:
      200:
        description: Returns an array of Detection models on the uploaded image
        schema:
          type: array
          items:
            $ref: "#/definitions/Detection"
      400:
        description: Invalid input parameter
      415:
        description: Unsupported media type in the request body. Currently only image/png, image/jpeg, image/tiff are supported.
    """
    ibs = current_app.ibs

    # Input argument validation
    image = microsoft_image_upload(ibs)
    microsoft_detect_input_validation(model, score_threshold, use_nms, nms_threshold)

    try:
        images = [image]
        detections_list = microsoft_detect(
            ibs, images, model, score_threshold, use_nms, nms_threshold
        )
    except Exception:
        logger.info(str(traceback.format_exc()))
        raise controller_inject.WebException(
            'Detection process failed for an unknown reason'
        )

    detections_list = detections_list.get('detections_list')
    assert len(detections_list) == 1
    detections = detections_list[0]

    logger.info(detections)
    return detections


@register_api(_prefix('detect/batch'), methods=['POST'])
def microsoft_detect_batch(
    ibs,
    images,
    model,
    score_threshold=0.0,
    use_nms=True,
    nms_threshold=0.4,
    asynchronous=True,
    callback_url=None,
    callback_method=None,
    *args,
    **kwargs,
):
    r"""
    The asynchronous variant of POST 'detect' that takes in a list of Image models and returns a task ID.

    It may be more ideal for a particular application to upload many images at one time and perform processing later in a large batch.  This type of batch detection is certainly much more efficient because the detection on GPU can process more images in parallel.  However, if you intend to run the detector on an upload as quickly as possible, please use the on-demand, non-batched API.
    ---
    parameters:
    - name: images
      in: body
      description: A JSON list of Image models for which detection will be processed.
      required: true
      type: array
    - name: model
      in: formData
      description: The model name to use with detection.  Available models can be queried using the API $PREFIX/detect/model/
      required: true
      type: string
    - name: score_threshold
      in: formData
      description: A score threshold to filter the detection results.  Default 0.0.  Must be in range [0, 1], inclusive
      required: false
      type: number
      format: float
    - name: use_nms
      in: formData
      description: Flag to turn on/off Non-Maximum Suppression (NMS).  Default True.
      required: false
      type: boolean
    - name: nms_threshold
      in: formData
      description: A Intersection-over-Union (IoU) threshold to suppress detection results with NMS.  Default 0.0.  Must be in range [0, 1], inclusive
      required: false
      type: number
      format: float
    - name: callback_url
      in: formData
      description: The URL of where to callback when the task is completed, must be a fully resolvable address and accessible.  The callback will include a 'body' parameter called `task` which will provide a Task model
      required: false
      type: string
      format: url
    - name: callback_method
      in: formData
      description: The HTTP method for which to make the callback.  Default POST.
      required: false
      type: string
      enum:
      - get
      - post
      - put
      - delete
    consumes:
    - multipart/form-data
    responses:
      200:
        description: Returns a Task model
        schema:
          $ref: "#/definitions/Task"
      400:
        description: Invalid input parameter
      x-task-response:
        description: The task returns an array of arrays of Detection models, in parallel lists with the provided Image models
        schema:
          type: array
          items:
            schema:
              type: array
              items:
                $ref: "#/definitions/Detection"
    """
    ibs = current_app.ibs

    # Input argument validation
    for index, image in enumerate(images):
        try:
            parameter = 'images:%d' % (index,)
            assert 'uuid' in image, 'Image Model provided is invalid, missing UUID key'
        except AssertionError as ex:
            raise controller_inject.WebInvalidInput(str(ex), parameter)

    microsoft_detect_input_validation(model, score_threshold, use_nms, nms_threshold)

    try:
        parameter = 'asynchronous'
        assert isinstance(asynchronous, bool), 'Asynchronous flag must be a boolean'

        parameter = 'callback_url'
        assert callback_url is None or isinstance(
            callback_url, str
        ), 'Callback URL must be a string'
        if callback_url is not None:
            assert callback_url.startswith('http://') or callback_url.startswith(
                'https://'
            ), 'Callback URL must start with http:// or https://'

        parameter = 'callback_method'
        assert callback_method is None or isinstance(
            callback_method, str
        ), 'Callback URL must be a string'
        if callback_method is not None:
            callback_method = callback_method.lower()
            assert callback_method in [
                'get',
                'post',
                'put',
                'delete',
            ], 'Unsupported callback method, must be one of ("get", "post", "put", "delete")'
    except AssertionError as ex:
        raise controller_inject.WebInvalidInput(str(ex), parameter)

    args = (
        images,
        model,
    )
    kwargs = {
        'score_threshold': score_threshold,
        'use_nms': use_nms,
        'nms_threshold': nms_threshold,
    }

    if asynchronous:
        taskid = ibs.job_manager.jobiface.queue_job(
            'microsoft_detect', callback_url, callback_method, *args, **kwargs
        )
        response = _task(ibs, taskid)
    else:
        response = ibs.microsoft_detect(*args, **kwargs)

    return response


@register_api(_prefix('identify'), methods=['POST'])
def microsoft_identify(
    ibs,
    query_annotation,
    database_annotations,
    algorithm,
    callback_url=None,
    callback_method=None,
    *args,
    **kwargs,
):
    r"""
    The asynchronous call to identify a list of pre-uploaded query annotations against a database of annotations.  Returns a task ID.

    This call expects all of the data has been already uploaded and the
    appropriate metadata set.  For example, this call expects any known
    ground-truth nameshave already been made and assigned to the appropriate
    Annotation models.  Conversely, we also provide a method for marking individual
    reviews between two Annotation models.  The decision between a given pair
    should always be one of ['match', 'nomatch', or 'notcomp'].  The full docuemntation
    for this call can be seen in the POST 'decision' API.
    ---
    parameters:
    - name: query_annotation
      in: formData
      description: An Annotation models to query for an identification
      required: true
      schema:
        $ref: "#/definitions/Annotation"
    - name: database_annotations
      in: formData
      description: A JSON list of Annotation models to compare the query Annotation against
      required: true
      type: array
      items:
        $ref: "#/definitions/Annotation"
    - name: algorithm
      in: formData
      description: The algorithm you with to run ID with.  Must be one of "HotSpotter", "CurvRank", "Finfindr", or "Deepsense"
      required: true
      type: string
    - name: callback_url
      in: formData
      description: The URL of where to callback when the task is completed, must be a fully resolvable address and accessible.  The callback will include a 'body' parameter called `task` which will provide a Task model
      required: false
      type: string
      format: url
    - name: callback_method
      in: formData
      description: The HTTP method for which to make the callback.  Default POST.
      required: false
      type: string
      enum:
      - get
      - post
      - put
      - delete
    consumes:
    - multipart/form-data
    responses:
      200:
        description: Returns a Task model
        schema:
          $ref: "#/definitions/Task"
      400:
        description: Invalid input parameter
    """
    qaid = _ensure_annotations(ibs, query_annotation)
    daid_list = _ensure_annotations(ibs, database_annotations)

    try:
        parameter = 'database_annotations'
        assert (
            len(daid_list) > 0
        ), 'Cannot specify an empty list of database Annotations to compare against'

        parameter = 'algorithm'
        assert isinstance(algorithm, str), 'Must specify the algorithm as a string'
        algorithm = algorithm.lower()
        assert algorithm in [
            'hotspotter',
            'curvrank',
            'deepsense',
            'finfindr',
            'kaggle7',
            'kaggleseven',
        ], 'Must specify the algorithm for ID as HotSpotter, CurvRank, Deepsense, Finfindr, Kaggle7'

        parameter = 'callback_url'
        assert callback_url is None or isinstance(
            callback_url, str
        ), 'Callback URL must be a string'
        if callback_url is not None:
            assert callback_url.startswith('http://') or callback_url.startswith(
                'https://'
            ), 'Callback URL must start with http:// or https://'

        parameter = 'callback_method'
        assert callback_method is None or isinstance(
            callback_method, str
        ), 'Callback URL must be a string'
        if callback_method is not None:
            callback_method = callback_method.lower()
            assert callback_method in [
                'get',
                'post',
                'put',
                'delete',
            ], 'Unsupported callback method, must be one of ("get", "post", "put", "delete")'
    except AssertionError as ex:
        raise controller_inject.WebInvalidInput(str(ex), parameter)

    if algorithm in ['hotspotter']:
        query_config_dict = {}
    elif algorithm in ['curvrank']:
        query_config_dict = {
            'pipeline_root': 'CurvRankFluke',
        }
    elif algorithm in ['deepsense']:
        query_config_dict = {
            'pipeline_root': 'Deepsense',
        }
    elif algorithm in ['finfindr']:
        query_config_dict = {
            'pipeline_root': 'Finfindr',
        }
    elif algorithm in ['kaggle7', 'kaggleseven']:
        query_config_dict = {
            'pipeline_root': 'KaggleSeven',
        }

    user_feedback = {
        'aid1': [],
        'aid2': [],
        'p_match': [],
        'p_nomatch': [],
        'p_notcomp': [],
    }
    echo_query_params = False
    args = (
        [qaid],
        daid_list,
        user_feedback,
        query_config_dict,
        echo_query_params,
    )
    kwargs = {
        'endpoint': url_for('microsoft_identify_visualize'),
        'n': 5,
    }
    taskid = ibs.job_manager.jobiface.queue_job(
        'query_chips_graph_microsoft', callback_url, callback_method, *args, **kwargs
    )

    return _task(ibs, taskid)


@register_ibs_method
def query_chips_graph_microsoft(ibs, *args, **kwargs):
    endpoint = kwargs.pop('endpoint')
    result = ibs.query_chips_graph(*args, **kwargs)

    cm_dict = result['cm_dict']
    cm_key_list = list(cm_dict.keys())
    assert len(cm_key_list) == 1
    cm_key = cm_key_list[0]
    cm = cm_dict[cm_key]

    qaid = ibs.get_annot_aids_from_uuid(cm['qannot_uuid'])
    reference = cm['dannot_extern_reference']

    response = {
        'annotations': [],
        'names': [],
    }

    zipped = list(
        zip(cm['annot_score_list'], cm['dannot_uuid_list'], cm['dannot_extern_list'])
    )
    zipped = sorted(zipped, reverse=True)
    for dannot_score, dannot_uuid, dannot_extern in zipped:
        daid = ibs.get_annot_aids_from_uuid(dannot_uuid)
        annotation = _annotation(ibs, daid)
        extern_url_postfix = None
        if dannot_extern:
            query_annotation_ = ut.to_json(_annotation(ibs, qaid)).replace(' ', '')
            database_annotation_ = ut.to_json(_annotation(ibs, daid)).replace(' ', '')
            args = (
                endpoint,
                reference,
                query_annotation_,
                database_annotation_,
            )
            extern_url_postfix = (
                '%s?reference=%s&query_annotation=%s&database_annotation=%s' % args
            )
        response['annotations'].append(
            {
                'score': dannot_score,
                'annotation': annotation,
                'visualize': extern_url_postfix,
            }
        )

    zipped = list(zip(cm['name_score_list'], cm['unique_name_list']))
    zipped = sorted(zipped, reverse=True)
    for name_score, name_text in zipped:
        if name_text == const.UNKNOWN:
            continue
        nid = ibs.get_name_rowids_from_text(name_text)
        response['names'].append({'score': name_score, 'name': _name(ibs, nid)})

    return response


@register_route(_prefix('visualize'), methods=['GET'])
def microsoft_identify_visualize(
    reference, query_annotation, database_annotation, version='heatmask'
):
    r"""
    Visualize the results of matching, precomputed and rendered to disk for easy look-up.

    ---
    parameters:
    - name: reference
      in: formData
      description: A reference string to refer to the rendered match results on disk for a given identification Task
      required: true
      type: string
    - name: query_annotation
      in: formData
      description: The query Annotation model
      required: true
      schema:
        $ref: "#/definitions/Annotation"
    - name: database_annotation
      in: formData
      description: The database Annotation model
      required: true
      schema:
        $ref: "#/definitions/Annotation"
    - name: version
      in: formData
      description: The version of the visualization.  Must be one of "heatmask" or "original".  Defaults to "heatmask"
      required: false
      type: string
    consumes:
    - multipart/form-data
    responses:
      200:
        description: Returns a Task model
        schema:
          $ref: "#/definitions/Task"
      400:
        description: Invalid input parameter
    """
    ibs = current_app.ibs

    from wbia.web.apis_query import query_chips_graph_match_thumb

    qaid = _ensure_annotations(ibs, query_annotation)
    daid = _ensure_annotations(ibs, database_annotation)
    quuid, duuid = ibs.get_annot_uuids([qaid, daid])
    quuid = str(quuid)
    duuid = str(duuid)
    response = query_chips_graph_match_thumb(
        extern_reference=reference,
        query_annot_uuid=quuid,
        database_annot_uuid=duuid,
        version=version,
    )
    return response


@register_api(_prefix('task'), methods=['GET'])
def microsoft_task_status(ibs, task):
    r"""
    Check the status of an asynchronous Task.

    A Task is an asynchronous task that was launched as a background process with an optional callback.  The status of a given Task with a UUID can be checked with this call.  The status of the call depends on where in the execution queue the Task is currently, which will be processed in a first-come-first-serve list and only one Task at a time to present atomicity of the API.

    The status can be one of the following:
    - received   -> The Task request was received but has not passed any input validation.
    - accepted   -> The Task request has passed basic input validation and will be queued soon for execution.
    - queued     -> The Task is queued in the execution list and will be processed in order and one at a time.
    - working    -> The Task is being processed, awaiting completed results or an error exception
    - publishing -> The Task is publishing the results of the background API call.
    - completed  -> One of two end states: the Task is complete, completed results available for downloading with the REST API.
    - exception  -> One of two end states: the Task has encountered an error, an error message can be received using the results REST API.
    - unknown    -> The Task you asked for is not known, indicating that the either UUID is not recognized (i.e. a Task with that ID was never currently created) or the server has been restarted.

    **Important: when the API server is restarted, all queued and running background Tasks are killed and all Task requests and cached results are deleted.**
    ---
    parameters:
    - name: task
      in: body
      description: A Task model
      required: true
      schema:
        $ref: "#/definitions/Task"
    responses:
      200:
        description: Returns the status of the provided Task
        schema:
          type: object
          properties:
            status:
              type: string
              enum:
              - received
              - accepted
              - queued
              - working
              - publishing
              - completed
              - exception
              - unknown
      400:
        description: Invalid input parameter
    """
    ibs = current_app.ibs

    # Input argument validation
    try:
        parameter = 'task'
        assert 'uuid' in task, 'Task Model provided is invalid, missing UUID key'
    except AssertionError as ex:
        raise controller_inject.WebInvalidInput(str(ex), parameter)

    uuid_ = task.get('uuid', None)
    assert uuid_ is not None
    status = ibs.get_job_status(uuid_)
    status = status.get('jobstatus', None)
    return {'status': status}


@register_api(_prefix('task'), methods=['POST'])
def microsoft_task_result(ibs, task):
    r"""
    Retrieve the result of a completed asynchronous Task
    ---
    parameters:
    - name: task
      in: body
      description: A Task model
      required: true
      schema:
        $ref: "#/definitions/Task"
    responses:
      200:
        description: Returns the result of the provided Task
      400:
        description: Invalid input parameter
    """
    ibs = current_app.ibs

    # Input argument validation
    try:
        parameter = 'task'
        assert 'uuid' in task, 'Task Model provided is invalid, missing UUID key'
    except AssertionError as ex:
        raise controller_inject.WebInvalidInput(str(ex), parameter)

    uuid_ = task.get('uuid', None)
    assert uuid_ is not None
    result = ibs.get_job_result(uuid_)
    result = result.get('json_result', None)
    return result


@register_api(_prefix('image'), methods=['DELETE'])
def microsoft_image_delete(ibs, image):
    r"""
    Delete an Image model from the database.

    Also deletes any derived Detection models or any added Annotations for the deleted Image model.
    ---
    parameters:
    - name: image
      in: body
      description: A Image model to delete
      required: true
      schema:
        $ref: "#/definitions/Image"
    produces:
    - application/json
    responses:
      200:
        description: Returns a success status flag (True if deleted)
        schema:
          type: boolean
      400:
        description: Invalid input parameter

    """
    ibs = current_app.ibs

    # Input argument validation
    gid = _ensure_images(ibs, image)
    ibs.delete_images(gid)

    return {'status': 'deleted'}


@register_api(_prefix('annotation'), methods=['DELETE'])
def microsoft_annotation_delete(ibs, annotation):
    r"""
    Delete an Annotation model from the database.
    ---
    parameters:
    - name: annotation
      in: body
      description: A Annotation model to delete
      required: true
      schema:
        $ref: "#/definitions/Annotation"
    produces:
    - application/json
    responses:
      200:
        description: Returns a success status flag (True if deleted)
        schema:
          type: boolean
      400:
        description: Invalid input parameter

    """
    ibs = current_app.ibs

    # Input argument validation
    aid = _ensure_annotations(ibs, annotation)
    ibs.delete_annots(aid)

    return {'status': 'deleted'}


@register_api(_prefix('name'), methods=['DELETE'])
def microsoft_name_delete(ibs, name):
    r"""
    Delete a Name model from the database.

    This operation DOES NOT delete any associated Annotation models with this Name, it simply disassociates them.
    ---
    parameters:
    - name: name
      in: body
      description: A Name model to delete
      required: true
      schema:
        $ref: "#/definitions/Name"
    produces:
    - application/json
    responses:
      200:
        description: Returns a success status flag (True if deleted)
        schema:
          type: boolean
      400:
        description: Invalid input parameter

    """
    ibs = current_app.ibs

    # Input argument validation
    nid = _ensure_names(ibs, name)
    ibs.delete_names(nid)

    return {'status': 'deleted'}


@register_api(_prefix('test/annotations'), methods=['GET'], __api_plural_check__=False)
def microsoft_get_test_data(ibs, dataset):
    r"""
    Return test data annotations.
    ---
    parameters:
    - name: dataset
      in: body
      description: A name for the dataset.  Must be one of 'zebra', 'dolphin', 'humpback'
      required: true
      type: string
    produces:
    - application/json
    responses:
      200:
        description: A JSON object with the query and database Annotation model lists
      400:
        description: Invalid input parameter
    """
    response = ibs.api_test_datasets_id(dataset)

    for key in response:
        aid_list = response[key]
        response[key] = [_annotation(ibs, aid) for aid in aid_list]

    return response



# -*- coding: utf-8 -*-
import utool as ut


def impaint_mask(img, label_colors=None, init_mask=None, init_label=None):
    r"""
    CommandLine:
        python -m wbia.plottool.interact_impaint --test-impaint_mask

    References:
        http://docs.opencv.org/trunk/doc/py_tutorials/py_gui/py_mouse_handling/py_mouse_handling.html

    TODO: Slider for transparency
    TODO: Label selector

    Example:
        >>> # DISABLE_DOCTEST
        >>> from wbia.plottool.interact_impaint import *  # NOQA
        >>> import utool as ut
        >>> import vtool as vt
        >>> img_fpath = ut.grab_test_imgpath('lena.png')
        >>> img = vt.imread(img_fpath)
        >>> label_colors = [255, 200, 100, 0]
        >>> result = impaint_mask(img, label_colors)
        >>> # verify results
        >>> print(result)
    """
    import cv2
    import numpy as np

    print('begining impaint mask. c=circle, r=rect')

    globals_ = dict(
        drawing=False,  # true if mouse is pressed
        mode='rect',  # if True, draw rectangle. Press 'm' to toggle to curve
        color=255,
        fgcolor=255,
        bgcolor=0,
        label_index=0,
        radius=25,
        transparency=0.25,
        ix=-1,
        iy=-1,
    )

    # mouse callback function
    def draw_shape(x, y):
        keys = ['mode', 'ix', 'iy', 'color', 'radius']
        mode, ix, iy, color, radius = ut.dict_take(globals_, keys)
        if mode == 'rect':
            cv2.rectangle(mask, (ix, iy), (x, y), color, -1)
        elif mode == 'circ':
            cv2.circle(mask, (x, y), radius, color, -1)

    def mouse_callback(event, x, y, flags, param):
        # keys =  ['drawing', 'mode', 'ix', 'iy', 'color']
        # drawing, mode, ix, iy, color = ut.dict_take(globals_, keys)

        if event in [cv2.EVENT_RBUTTONDOWN, cv2.EVENT_LBUTTONDOWN]:
            globals_['drawing'] = True
            globals_['ix'], globals_['iy'] = x, y
            if event == cv2.EVENT_RBUTTONDOWN:
                globals_['color'] = globals_['bgcolor']
            elif event == cv2.EVENT_LBUTTONDOWN:
                globals_['color'] = globals_['fgcolor']
        elif event == cv2.EVENT_MOUSEMOVE:
            if globals_['drawing'] is True:
                draw_shape(x, y)
        elif event in [cv2.EVENT_LBUTTONUP, cv2.EVENT_RBUTTONUP]:
            globals_['drawing'] = False
            draw_shape(x, y)
            if event == cv2.EVENT_RBUTTONUP:
                globals_['color'] = globals_['fgcolor']
            elif event == cv2.EVENT_LBUTTONUP:
                pass
                # globals_['color'] = 255

    if label_colors is None:
        color_list = [255, 0]
    else:
        color_list = label_colors[:]

    # Choose colors/labels to start with
    if init_label is None:
        init_color = 0
    else:
        init_color = color_list[init_label]

    print('color_list = %r' % (color_list,))
    print('init_color=%r' % (init_color,))

    title = 'masking image'
    if init_mask is not None:
        try:
            mask = init_mask[:, :, 0].copy()
        except Exception:
            mask = init_mask.copy()
    else:
        mask = np.zeros(img.shape[0:2], np.uint8) + init_color
    transparent_mask = np.zeros(img.shape[0:2], np.float32)
    cv2.namedWindow(title)
    cv2.setMouseCallback(title, mouse_callback)

    print('Valid Keys: r,c,t,l,q')
    while 1:
        # Blend images
        transparency = globals_['transparency']
        # Move from 0 to 1
        np.divide(mask, 255.0, out=transparent_mask)
        # Unmask room for a bit of transparency
        np.multiply(transparent_mask, (1.0 - transparency), out=transparent_mask)
        # Add a bit of transparency
        np.add(transparent_mask, transparency, out=transparent_mask)
        # Multiply the image by the transparency mask
        masked_image = (img * transparent_mask[:, :, None]).astype(np.uint8)
        cv2.imshow(title, masked_image)
        keycode = cv2.waitKey(1) & 0xFF
        if keycode == ord('r'):
            globals_['mode'] = 'rect'
        if keycode == ord('c'):
            globals_['mode'] = 'circ'
        if keycode == ord('t'):
            globals_['transparency'] = (globals_['transparency'] + 0.25) % 1.0
        if keycode == ord('l'):
            globals_['label_index'] = (globals_['label_index'] + 1) % len(color_list)
            globals_['fgcolor'] = color_list[globals_['label_index']]
            print('fgcolor = %r' % (globals_['fgcolor'],))
        if keycode == ord('q') or keycode == 27:
            break

    cv2.destroyAllWindows()
    return mask


def cached_impaint(
    bgr_img,
    cached_mask_fpath=None,
    label_colors=None,
    init_mask=None,
    aug=False,
    refine=False,
):
    import vtool as vt

    if cached_mask_fpath is None:
        cached_mask_fpath = 'image_' + ut.hashstr_arr(bgr_img) + '.png'
    if aug:
        cached_mask_fpath += '.' + ut.hashstr_arr(bgr_img)
        if label_colors is not None:
            cached_mask_fpath += ut.hashstr_arr(label_colors)
        cached_mask_fpath += '.png'
    # cached_mask_fpath = 'tmp_mask.png'
    if refine or not ut.checkpath(cached_mask_fpath):
        if refine and ut.checkpath(cached_mask_fpath):
            if init_mask is None:
                init_mask = vt.imread(cached_mask_fpath, grayscale=True)
        custom_mask = impaint_mask(
            bgr_img, label_colors=label_colors, init_mask=init_mask
        )
        vt.imwrite(cached_mask_fpath, custom_mask)
    else:
        custom_mask = vt.imread(cached_mask_fpath, grayscale=True)
    return custom_mask


def demo():
    r"""
    CommandLine:
        python -m wbia.plottool.interact_impaint --test-demo

    References:
        http://docs.opencv.org/trunk/doc/py_tutorials/py_gui/py_mouse_handling/py_mouse_handling.html

    Example:
        >>> # DISABLE_DOCTEST
        >>> from wbia.plottool.interact_impaint import *  # NOQA
        >>> # build test data
        >>> # execute function
        >>> result = demo()
        >>> # verify results
        >>> print(result)
    """
    import cv2
    import numpy as np

    globals_ = dict(
        drawing=False,  # true if mouse is pressed
        mode=False,  # if True, draw rectangle. Press 'm' to toggle to curve
        ix=-1,
        iy=-1,
    )

    # mouse callback function
    def draw_circle(event, x, y, flags, param):
        if event == cv2.EVENT_LBUTTONDOWN:
            globals_['drawing'] = True
            globals_['ix'], globals_['iy'] = x, y

        elif event == cv2.EVENT_MOUSEMOVE:
            if globals_['drawing'] is True:
                if globals_['mode'] is True:
                    cv2.rectangle(
                        img, (globals_['ix'], globals_['iy']), (x, y), (0, 255, 0), -1
                    )
                else:
                    cv2.circle(img, (x, y), 5, (0, 0, 255), -1)

        elif event == cv2.EVENT_LBUTTONUP:
            globals_['drawing'] = False
            if globals_['mode'] is True:
                cv2.rectangle(
                    img, (globals_['ix'], globals_['iy']), (x, y), (0, 255, 0), -1
                )
            else:
                cv2.circle(img, (x, y), 5, (0, 0, 255), -1)

    img = np.zeros((512, 512, 3), np.uint8)
    cv2.namedWindow('image')
    cv2.setMouseCallback('image', draw_circle)

    while 1:
        cv2.imshow('image', img)
        keycode = cv2.waitKey(1) & 0xFF
        if keycode == ord('m'):
            globals_['mode'] = not globals_['mode']
        elif keycode == 27:
            break

    cv2.destroyAllWindows()



# -*- coding: utf-8 -*-
import logging
import utool as ut
import networkx as nx
import itertools as it
from wbia.algo.graph.nx_utils import edges_inside, e_

print, rrr, profile = ut.inject2(__name__)
logger = logging.getLogger('wbia')


class GraphHelperMixin(ut.NiceRepr):
    def __nice__(self):
        return 'nNodes={}, nEdges={}'.format(
            self.number_of_nodes(), self.number_of_edges()
        )

    def has_nodes(self, nodes):
        return (self.has_node(node) for node in nodes)

    def has_edges(self, edges):
        return (self.has_edge(*edge) for edge in edges)

    def edges(self, nbunch=None, data=False, default=None):
        # Force edges to always be returned in upper triangular form
        edges = super(GraphHelperMixin, self).edges(nbunch, data, default)
        if data:
            return (e_(u, v) + (d,) for u, v, d in edges)
        else:
            return (e_(u, v) for u, v in edges)


class NiceGraph(nx.Graph, GraphHelperMixin):
    pass


class nx_UnionFind(object):
    """
    Based of nx code
    """

    def __init__(self, elements=None):
        if elements is None:
            elements = ()
        self.parents = {}
        self.weights = {}
        self.add_elements(elements)

    def clear(self):
        self.parents = {}
        self.weights = {}

    def __getitem__(self, element):
        # check for previously unknown element
        if self.add_element(element):
            return element
        # find path of objects leading to the root
        path = [element]
        root = self.parents[element]
        while root != path[-1]:
            path.append(root)
            root = self.parents[root]
        # compress the path and return
        for ancestor in path:
            self.parents[ancestor] = root
        return root

    def __iter__(self):
        return iter(self.parents)

    def rebalance(self, elements=None):
        if elements is None:
            elements = list(self.parents.keys())
        # Make sure only one operation is needed to lookup any node
        for x in elements:
            parent = self[x]
            self.parents[x] = parent
            self.weights[x] = 1

    def to_sets(self):
        for block in it.groups(self.parents).values():
            yield block

    def union(self, *objects):
        """Find the sets containing the objects and merge them all."""
        roots = [self[x] for x in objects]
        # HACK: use the lowest node number to preserve
        # node labels through cuts. (has some runtime penalty)
        if False:
            # Find the best root with the maximum weight
            best = max(roots, key=lambda r: self.weights[r])
        else:
            best = min(roots)
        for r in roots:
            if r != best:
                self.weights[best] += self.weights[r]
                self.parents[r] = best

    def remove_entire_cc(self, elements):
        # NOTE: this will not work in general. This only
        # works if all elements are a unique component.
        for x in elements:
            del self.weights[x]
            del self.parents[x]

    def add_element(self, x):
        if x not in self.parents:
            self.weights[x] = 1
            self.parents[x] = x
            return True
        return False

    def add_elements(self, elements):
        for x in elements:
            if x not in self.parents:
                self.weights[x] = 1
                self.parents[x] = x


class DynConnGraph(nx.Graph, GraphHelperMixin):
    """
    Dynamically connected graph.

    Maintains a data structure parallel to a normal networkx graph that
    maintains dynamic connectivity for fast connected compoment queries.

    Underlying Data Structures and limitations are

    Data Structure     | Insertion | Deletion | CC Find |
    -----------------------------------------------------
    * UnionFind        | lg(n)     |    n     |  No
    * UnionFind2       |    n*     |    n     |  1
    * EulerTourForest  | lg^2(n)   | lg^2(n)  |  lg(n) / lglg(n) - - Ammortized

    * it seems to be very quick

    References:
        https://courses.csail.mit.edu/6.851/spring14/lectures/L20.pdf
        https://courses.csail.mit.edu/6.851/spring14/lectures/L20.html
        http://cs.stackexchange.com/questions/33595/maintaining-connecte
        https://en.wikipedia.org/wiki/Dynamic_connectivity#Fully_dynamic_connectivity

    CommandLine:
        python -m wbia.algo.graph.nx_dynamic_graph DynConnGraph

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.algo.graph.nx_dynamic_graph import *  # NOQA
        >>> self = DynConnGraph()
        >>> self.add_edges_from([(1, 2), (2, 3), (4, 5), (6, 7), (7, 4)])
        >>> self.add_edges_from([(10, 20), (20, 30), (40, 50), (60, 70), (70, 40)])
        >>> self._ccs
        >>> u, v = 20, 1
        >>> assert self.node_label(u) != self.node_label(v)
        >>> assert self.connected_to(u) != self.connected_to(v)
        >>> self.add_edge(u, v)
        >>> assert self.node_label(u) == self.node_label(v)
        >>> assert self.connected_to(u) == self.connected_to(v)
        >>> self.remove_edge(u, v)
        >>> assert self.node_label(u) != self.node_label(v)
        >>> assert self.connected_to(u) != self.connected_to(v)
        >>> ccs = list(self.connected_components())
        >>> ut.quit_if_noshow()
        >>> import wbia.plottool as pt
        >>> pt.qtensure()
        >>> pt.show_nx(self)

    # todo: check if nodes exist when adding
    """

    def __init__(self, *args, **kwargs):
        # raise NotImplementedError('unfinished')
        self._ccs = {}
        self._union_find = nx_UnionFind()
        super(DynConnGraph, self).__init__(*args, **kwargs)

    def clear(self):
        super(DynConnGraph, self).clear()
        self._ccs = {}
        self._union_find.clear()

    def __nice__(self):
        return 'nNodes={}, nEdges={}, nCCs={}'.format(
            self.number_of_nodes(), self.number_of_edges(), self.number_of_components()
        )

    def number_of_components(self):
        return len(self._ccs)

    def component(self, label):
        return self._ccs[label]

    component_nodes = component

    def connected_to(self, node):
        return self._ccs[self._union_find[node]]

    def node_label(self, node):
        """
        Example:
            >>> # ENABLE_DOCTEST
            >>> from wbia.algo.graph.nx_dynamic_graph import *  # NOQA
            >>> self = DynConnGraph()
            >>> self.add_edges_from([(1, 2), (2, 3), (4, 5), (6, 7)])
            >>> assert self.node_label(2) == self.node_label(1)
            >>> assert self.node_label(2) != self.node_label(4)
        """
        return self._union_find[node]

    def node_labels(self, *nodes):
        return [self._union_find[node] for node in nodes]

    def are_nodes_connected(self, u, v):
        return ut.allsame(self.node_labels(u, v))

    def connected_components(self):
        """
        Example:
            >>> # ENABLE_DOCTEST
            >>> from wbia.algo.graph.nx_dynamic_graph import *  # NOQA
            >>> self = DynConnGraph()
            >>> self.add_edges_from([(1, 2), (2, 3), (4, 5), (6, 7)])
            >>> ccs = list(self.connected_components())
            >>> result = 'ccs = {}'.format(ut.repr2(ccs, nl=0))
            >>> print(result)
            ccs = [{1, 2, 3}, {4, 5}, {6, 7}]
        """
        for cc in self._ccs.values():
            yield cc

    def component_labels(self):
        for label in self._ccs.keys():
            yield label

    # -----

    def _cut(self, u, v):
        """ Decremental connectivity (slow) """
        old_nid1 = self._union_find[u]
        old_nid2 = self._union_find[v]
        if old_nid1 != old_nid2:
            return
        # Need to break appart entire component and then reconstruct it
        old_cc = self._ccs[old_nid1]
        del self._ccs[old_nid1]
        self._union_find.remove_entire_cc(old_cc)
        # Might be faster to just do DFS to find the CC
        internal_edges = edges_inside(self, old_cc)
        # Add nodes in case there are no edges to it
        for n in old_cc:
            self._add_node(n)
        for edge in internal_edges:
            self._union(*edge)

    def _union(self, u, v):
        """ Incremental connectivity (fast) """
        # logger.info('Union ({})'.format((u, v)))
        self._add_node(u)
        self._add_node(v)
        old_nid1 = self._union_find[u]
        old_nid2 = self._union_find[v]
        self._union_find.union(u, v)
        new_nid = self._union_find[u]
        for old_nid in [old_nid1, old_nid2]:
            if new_nid != old_nid:
                parts = self._ccs.pop(old_nid)
                # FIXME: this step can be quite bad for time complexity.
                # An Euler Tour Tree might solve the issue
                self._ccs[new_nid].update(parts)

    def _add_node(self, n):
        if self._union_find.add_element(n):
            # logger.info('Add ({})'.format((n)))
            self._ccs[n] = {n}

    def _remove_node(self, n):
        if n in self._union_find.parents:
            del self._union_find.weights[n]
            del self._union_find.parents[n]
            del self._ccs[n]

    def add_edge(self, u, v, **attr):
        """
        Example:
            >>> # ENABLE_DOCTEST
            >>> from wbia.algo.graph.nx_dynamic_graph import *  # NOQA
            >>> self = DynConnGraph()
            >>> self.add_edges_from([(1, 2), (2, 3), (4, 5), (6, 7), (7, 4)])
            >>> assert self._ccs == {1: {1, 2, 3}, 4: {4, 5, 6, 7}}
            >>> self.add_edge(1, 5)
            >>> assert self._ccs == {1: {1, 2, 3, 4, 5, 6, 7}}
        """
        self._union(u, v)
        super(DynConnGraph, self).add_edge(u, v, **attr)

    def add_edges_from(self, ebunch, **attr):
        ebunch = list(ebunch)
        # logger.info('add_edges_from %r' % (ebunch,))
        for e in ebunch:
            self._union(*e)
        super(DynConnGraph, self).add_edges_from(ebunch, **attr)

    # ----

    def add_node(self, n, **attr):
        self._add_node(n)
        super(DynConnGraph, self).add_node(n, **attr)

    def add_nodes_from(self, nodes, **attr):
        nodes = list(nodes)
        for n in nodes:
            self._add_node(n)
        super(DynConnGraph, self).add_nodes_from(nodes, **attr)

    # ----

    def remove_edge(self, u, v):
        """
        Example:
            >>> # ENABLE_DOCTEST
            >>> from wbia.algo.graph.nx_dynamic_graph import *  # NOQA
            >>> self = DynConnGraph()
            >>> self.add_edges_from([(1, 2), (2, 3), (4, 5), (6, 7), (7, 4)])
            >>> assert self._ccs == {1: {1, 2, 3}, 4: {4, 5, 6, 7}}
            >>> self.add_edge(1, 5)
            >>> assert self._ccs == {1: {1, 2, 3, 4, 5, 6, 7}}
            >>> self.remove_edge(1, 5)
            >>> assert self._ccs == {1: {1, 2, 3}, 4: {4, 5, 6, 7}}
        """
        super(DynConnGraph, self).remove_edge(u, v)
        self._cut(u, v)

    def remove_edges_from(self, ebunch):
        ebunch = list(ebunch)
        super(DynConnGraph, self).remove_edges_from(ebunch)
        # Can do this more efficiently for bulk edges
        for e in ebunch:
            self._cut(*e)

    # -----

    def remove_nodes_from(self, nodes):
        # remove edges as well
        nodes = list(nodes)
        for n in nodes:
            nbrs = list(self.adj[n].keys())
            self.remove_edges_from((n, v) for v in nbrs)
        for n in nodes:
            self._remove_node(n)
        super(DynConnGraph, self).remove_nodes_from(nodes)

    def remove_node(self, n):
        r"""
        CommandLine:
            python -m wbia.algo.graph.nx_dynamic_graph remove_node

        Example:
            >>> # ENABLE_DOCTEST
            >>> from wbia.algo.graph.nx_dynamic_graph import *  # NOQA
            >>> self = DynConnGraph()
            >>> self.add_edges_from([(1, 2), (2, 3), (4, 5), (5, 6), (6, 7), (7, 8), (8, 9)])
            >>> assert self._ccs == {1: {1, 2, 3}, 4: {4, 5, 6, 7, 8, 9}}
            >>> self.remove_node(2)
            >>> assert self._ccs == {1: {1}, 3: {3}, 4: {4, 5, 6, 7, 8, 9}}
            >>> self.remove_node(7)
            >>> assert self._ccs == {1: {1}, 3: {3}, 4: {4, 5, 6}, 8: {8, 9}}
        """
        # remove edges as well
        nbrs = list(self.adj[n].keys())
        self.remove_edges_from((n, v) for v in nbrs)
        self._remove_node(n)
        super(DynConnGraph, self).remove_node(n)

    @profile
    def subgraph(self, nbunch, dynamic=False):
        if dynamic is False:
            H = nx.Graph()
            nbunch = set(nbunch)
            H.add_nodes_from(nbunch)
            H.add_edges_from(edges_inside(self, nbunch))
        else:
            H = super(DynConnGraph, self).subgraph(nbunch)
            for n in nbunch:
                # need to add individual nodes
                H._add_node(n)
            # Recreate the connected compoment structure
            for u, v in H.edges():
                H._union(u, v)
        return H



# -*- coding: utf-8 -*-
import utool as ut
import six

# import itertools
from . import draw_func2 as df2
from wbia.plottool import plot_helpers as ph
from wbia.plottool import custom_constants
from wbia.plottool import custom_figure

# (print, print_, printDBG, rrr, profile) = ut.inject(
#    __name__, '[viz_featrow]', DEBUG=False)
ut.noinject(__name__, '[viz_featrow]')


def precisionstr(c='E', pr=2):
    return '%.' + str(pr) + c


def formatdist(val):
    pr = 3
    if val > 1000:
        return precisionstr('E', pr) % val
    elif val > 0.01 or val == 0:
        return precisionstr('f', pr) % val
    else:
        return precisionstr('e', pr) % val


# @ut.indent_func
def draw_feat_row(
    chip,
    fx,
    kp,
    sift,
    fnum,
    nRows,
    nCols=None,
    px=None,
    prevsift=None,
    origsift=None,
    aid=None,
    info='',
    type_=None,
    shape_labels=False,
    vecfield=False,
    multicolored_arms=False,
    draw_chip=False,
    draw_warped=True,
    draw_unwarped=True,
    draw_desc=True,
    rect=True,
    ori=True,
    pts=False,
    **kwargs
):
    """
    draw_feat_row

    SeeAlso:
        wbia.viz.viz_nearest_descriptors
        ~/code/wbia/wbia/viz/viz_nearest_descriptors.py

    CommandLine:

        # Use this to find the fx you want to visualize
        python -m wbia.plottool.interact_keypoints --test-ishow_keypoints --show --fname zebra.png

        # Use this to visualize the featrow
        python -m wbia.plottool.viz_featrow --test-draw_feat_row --show
        python -m wbia.plottool.viz_featrow --test-draw_feat_row --show --fname zebra.png --fx=121 --feat-all --no-sift
        python -m wbia.plottool.viz_featrow --test-draw_feat_row --dpath figures --save ~/latex/crall-candidacy-2015/figures/viz_featrow.jpg

    Example:
        >>> # DISABLE_DOCTEST
        >>> from wbia.plottool.viz_featrow import *  # NOQA
        >>> import wbia.plottool as pt
        >>> # build test data
        >>> kpts, vecs, imgBGR = pt.viz_keypoints.testdata_kpts()
        >>> chip = imgBGR
        >>> print('There are %d features' % (len(vecs)))
        >>> fx = ut.get_argval('--fx', type_=int, default=0)
        >>> kp = kpts[fx]
        >>> sift = vecs[fx]
        >>> fnum = 1
        >>> nRows = 1
        >>> nCols = 2
        >>> px = 0
        >>> if True:
        >>>     from wbia.scripts.thesis import TMP_RC
        >>>     import matplotlib as mpl
        >>>     mpl.rcParams.update(TMP_RC)
        >>> hack = ut.get_argflag('--feat-all')
        >>> sift = sift if not ut.get_argflag('--no-sift') else None
        >>> draw_desc = sift is not None
        >>> kw = dict(
        >>>     prevsift=None, origsift=None, aid=None, info='', type_=None,
        >>>     shape_labels=False, vecfield=False, multicolored_arms=True,
        >>>     draw_chip=hack, draw_unwarped=hack, draw_warped=True, draw_desc=draw_desc
        >>> )
        >>> # execute function
        >>> result = draw_feat_row(chip, fx, kp, sift, fnum, nRows, nCols, px,
        >>>                           rect=False, ori=False, pts=False, **kw)
        >>> # verify results
        >>> print(result)
        >>> pt.show_if_requested()
    """
    import numpy as np
    import vtool as vt

    # should not need ncols here

    if nCols is not None:
        if ut.VERBOSE:
            print('Warning nCols is no longer needed')
    # assert nCols_ == nCols
    nCols = draw_chip + draw_unwarped + draw_warped + draw_desc

    pnum_ = df2.make_pnum_nextgen(nRows, nCols, start=px)

    # pnum_ = df2.get_pnum_func(nRows, nCols, base=1)
    # countgen = itertools.count(1)

    # pnumgen_ = df2.make_pnum_nextgen(nRows, nCols, base=1)

    def _draw_patch(**kwargs):
        return df2.draw_keypoint_patch(
            chip,
            kp,
            sift,
            rect=rect,
            ori=ori,
            pts=pts,
            ori_color=custom_constants.DEEP_PINK,
            multicolored_arms=multicolored_arms,
            **kwargs
        )

    # Feature strings
    xy_str, shape_str, scale, ori_str = ph.kp_info(kp)

    if draw_chip:
        pnum = pnum_()
        df2.imshow(chip, fnum=fnum, pnum=pnum)
        kpts_kw = dict(ell_linewidth=1, ell_alpha=1.0)
        kpts_kw.update(kwargs)
        df2.draw_kpts2([kp], **kpts_kw)

    if draw_unwarped:
        # Draw the unwarped selected feature
        # ax = _draw_patch(fnum=fnum, pnum=pnum_(px + six.next(countgen)))
        # pnum = pnum_(px + six.next(countgen)
        pnum = pnum_()
        ax = _draw_patch(fnum=fnum, pnum=pnum)
        ph.set_plotdat(ax, 'viztype', 'unwarped')
        ph.set_plotdat(ax, 'aid', aid)
        ph.set_plotdat(ax, 'fx', fx)
        if shape_labels:
            unwarped_lbl = 'affine feature invV =\n' + shape_str + '\n' + ori_str
            custom_figure.set_xlabel(unwarped_lbl, ax)

    if draw_warped:
        # Draw the warped selected feature
        # ax = _draw_patch(fnum=fnum, pnum=pnum_(px + six.next(countgen)), warped=True)
        pnum = pnum_()
        ax = _draw_patch(fnum=fnum, pnum=pnum, warped=True, **kwargs)
        ph.set_plotdat(ax, 'viztype', 'warped')
        ph.set_plotdat(ax, 'aid', aid)
        ph.set_plotdat(ax, 'fx', fx)
        if shape_labels:
            warped_lbl = ('warped feature\n' + 'fx=%r scale=%.1f\n' + '%s') % (
                fx,
                scale,
                xy_str,
            )
        else:
            warped_lbl = ''
        warped_lbl += info
        custom_figure.set_xlabel(warped_lbl, ax)

    if draw_desc:
        border_color = {
            'None': None,
            'query': None,
            'match': custom_constants.BLUE,
            'norm': custom_constants.ORANGE,
        }.get(str(type_).lower(), None)
        if border_color is not None:
            df2.draw_border(ax, color=border_color)

        # Draw the SIFT representation
        # pnum = pnum_(px + six.next(countgen))
        pnum = pnum_()
        sift_as_vecfield = ph.SIFT_OR_VECFIELD or vecfield
        if sift_as_vecfield:
            custom_figure.figure(fnum=fnum, pnum=pnum)
            df2.draw_keypoint_gradient_orientations(chip, kp, sift=sift)
        else:
            if sift.dtype.type == np.uint8:
                # sigtitle =  'sift histogram' if (px % 3) == 0 else ''
                # ax = df2.plot_sift_signature(sift, sigtitle, fnum=fnum, pnum=pnum)
                ax = df2.plot_sift_signature(sift, fnum=fnum, pnum=pnum)
            else:
                # sigtitle =  'descriptor vector' if (px % 3) == 0 else ''
                ax = df2.plot_descriptor_signature(sift, fnum=fnum, pnum=pnum)
            ax._hs_viztype = 'histogram'
        # dist_list = ['L1', 'L2', 'hist_isect', 'emd']
        # dist_list = ['L2', 'hist_isect']
        # dist_list = ['L2']
        # dist_list = ['bar_L2_sift', 'cos_sift']
        # dist_list = ['L2_sift', 'bar_cos_sift']
        dist_list = ['L2_sift']
        dist_str_list = []
        if origsift is not None:
            distmap_orig = vt.compute_distances(sift, origsift, dist_list)
            dist_str_list.append(
                'query_dist: '
                + ', '.join(
                    [
                        '(%s, %s)' % (key, formatdist(val))
                        for key, val in six.iteritems(distmap_orig)
                    ]
                )
            )
        if prevsift is not None:
            distmap_prev = vt.compute_distances(sift, prevsift, dist_list)
            dist_str_list.append(
                'prev_dist: '
                + ', '.join(
                    [
                        '(%s, %s)' % (key, formatdist(val))
                        for key, val in six.iteritems(distmap_prev)
                    ]
                )
            )
        dist_str = '\n'.join(dist_str_list)
        custom_figure.set_xlabel(dist_str)
    return px + nCols



# -*- coding: utf-8 -*-
import logging
import utool as ut
import six
import itertools as it
from wbia import _wbia_object
from wbia.control.controller_inject import make_ibs_register_decorator

(print, rrr, profile) = ut.inject2(__name__, '[annot]')
logger = logging.getLogger('wbia')

CLASS_INJECT_KEY, register_ibs_method = make_ibs_register_decorator(__name__)


BASE_TYPE = type


@register_ibs_method
def annots(ibs, aids=None, uuids=None, **kwargs):
    """ Makes an Annots object """
    if uuids is not None:
        assert aids is None, 'specify one primary key'
        aids = ibs.get_annot_aids_from_uuid(uuids)
    if aids is None:
        aids = ibs.get_valid_aids()
    elif aids.__class__.__name__ == 'Annots':
        return aids
    aids = ut.ensure_iterable(aids)
    return Annots(aids, ibs, **kwargs)


@register_ibs_method
def matches(ibs, ams=None, edges=None, uuid_edges=None, **kwargs):
    """ Makes an Annots object """
    if uuid_edges is not None:
        assert ams is None, 'specify one primary key'
        assert edges is None, 'specify one primary key'
        uuids1, uuids2 = list(zip(*uuid_edges))
        aids1 = ibs.get_annot_aids_from_uuid(uuids1)
        aids2 = ibs.get_annot_aids_from_uuid(uuids2)
        ams = ibs.get_annotmatch_rowid_from_undirected_superkey(aids1, aids2)
    if edges is not None:
        assert ams is None, 'specify one primary key'
        assert uuid_edges is None, 'specify one primary key'
        aids1, aids2 = list(zip(*edges))
        ams = ibs.get_annotmatch_rowid_from_undirected_superkey(aids1, aids2)
    if ams is None:
        ams = ibs._get_all_annotmatch_rowids()
    elif ams.__class__.__name__ == 'AnnotMatches':
        return ams
    ams = ut.ensure_iterable(ams)
    return AnnotMatches(ams, ibs, **kwargs)


@register_ibs_method
def _annot_groups(ibs, aids_list=None, config=None):
    annots_list = [ibs.annots(aids, config=config) for aids in aids_list]
    return AnnotGroups(annots_list, ibs)


ANNOT_BASE_ATTRS = [
    'aid',
    'parent_aid',
    'multiple',
    'age_months_est_max',
    'age_months_est_min',
    'sex',
    'sex_texts',
    'uuids',
    'hashid_uuid',
    'visual_uuids',
    'hashid_visual_uuid',
    'semantic_uuids',
    'hashid_semantic_uuid',
    'verts',
    'thetas',
    'bboxes',
    'bbox_area',
    'species_uuids',
    'species',
    'species_rowids',
    'species_texts',
    'viewpoint_int',
    'viewpoint_code',
    'qualities',
    'quality_texts',
    'exemplar_flags',
    # DEPRICATE YAW
    'yaw_texts',
    'yaws',
    'yaws_asfloat',
    # Images
    # 'image_rowids',
    'gids',
    'image_uuids',
    'image_gps',
    'image_gps2',
    'image_unixtimes_asfloat',
    'image_datetime_str',
    'image_contributor_tag',
    # Names
    'nids',
    'names',
    'name_uuids',
    # Inferred from context attrs
    'contact_aids',
    'num_contact_aids',
    'groundfalse',
    'groundtruth',
    'num_groundtruth',
    'has_groundtruth',
    'otherimage_aids',
    # Image Set
    'imgset_uuids',
    'imgsetids',
    'image_set_texts',
    # Occurrence / Encounter
    'static_encounter',
    'encounter_text',
    'occurrence_text',
    'primary_imageset',
    # Tags
    'all_tags',
    'case_tags',
    'annotmatch_tags',
    'notes',
    # Processing State
    'reviewed',
    'reviewed_matching_aids',
    'has_reviewed_matching_aids',
    'num_reviewed_matching_aids',
    'detect_confidence',
]

ANNOT_SETTABLE_ATTRS = [
    'age_months_est_max',
    'age_months_est_min',
    'bboxes',
    'thetas',
    'verts',
    'qualities',
    'quality_texts',
    'viewpoint_int',
    'viewpoint_code',
    # DEPRICATE YAW
    'yaw_texts',
    'yaws',
    'sex',
    'sex_texts',
    'species',
    'exemplar_flags',
    'static_encounter',
    'multiple',
    'case_tags',
    'detect_confidence',
    'reviewed',
    'name_texts',
    'names',
    'notes',
    'parent_rowid',
]


class _AnnotPropInjector(BASE_TYPE):
    """
    Ignore:
        >>> from wbia import _wbia_object
        >>> import wbia
        >>> ibs = wbia.opendb(defaultdb='testdb1')
        >>> objname = 'annot'
        >>> blacklist = ['annot_pair']
        >>> _wbia_object._find_wbia_attrs(ibs, objname, blacklist)
    """

    def __init__(metaself, name, bases, dct):
        super(_AnnotPropInjector, metaself).__init__(name, bases, dct)
        metaself.rrr = rrr

        attrs = ANNOT_BASE_ATTRS

        settable_attrs = ANNOT_SETTABLE_ATTRS

        configurable_attrs = [
            # Chip
            'chip_dlensqrd',
            'chip_fpath',
            'chip_sizes',
            'chip_thumbpath',
            'chip_thumbtup',
            'chips',
            # Feat / FeatWeight / Kpts / Desc
            'feat_rowids',
            'num_feats',
            'featweight_rowids',
            'fgweights',
            'fgweights_subset',
            'kpts',
            'kpts_distinctiveness',
            'vecs',
            'vecs_cache',
            'vecs_subset',
        ]
        # misc = [
        #    'gar_rowids', 'alrids', 'alrids_oftype', 'lblannot_rowids',
        #    'lblannot_rowids_oftype', 'lblannot_value_of_lbltype', 'rows',
        #    'instancelist', 'lazy_dict', 'lazy_dict2', 'missing_uuid',
        #    'been_adjusted', 'class_labels',
        # ]
        # extra_attrs = [
        #    # Age / Sex
        #    'age_months_est', 'age_months_est_max', 'age_months_est_max_texts',
        #    'age_months_est_min', 'age_months_est_min_texts',
        #    'age_months_est_texts', 'sex', 'sex_texts',

        #    # Stats
        #    'stats_dict', 'per_name_stats', 'qual_stats', 'info', 'yaw_stats',
        #    'intermediate_viewpoint_stats',
        # ]
        # inverse_attrs = [
        #    # External lookups via superkeys
        #    'aids_from_semantic_uuid',
        #    'aids_from_uuid',
        #    'aids_from_visual_uuid',
        #    'rowids_from_partial_vuuids',
        # ]

        depcache_attrs = [
            ('hog', 'hog'),
            ('probchip', 'img'),
        ]

        aliased_attrs = {
            'time': 'image_unixtimes_asfloat',
            'gps': 'image_gps2',
            'chip_size': 'chip_sizes',
            'yaw': 'yaws_asfloat',
            'qual': 'qualities',
            'name': 'names',
            'nid': 'nids',
            'unary_tags': 'case_tags',
            # DEPRICATE
            'rchip': 'chips',
            'rchip_fpath': 'chip_fpath',
        }

        objname = 'annot'
        _wbia_object._inject_getter_attrs(
            metaself,
            objname,
            attrs,
            configurable_attrs,
            'depc_annot',
            depcache_attrs,
            settable_attrs,
            aliased_attrs,
        )

        # TODO: incorporate dynamic setters
        # def set_case_tags(self, tags):
        #    self._ibs.append_annot_case_tags(self._rowids, tags)
        # fget = metaself.case_tags.fget
        # fset = set_case_tags
        # setattr(metaself, 'case_tags', property(fget, fset))


try:
    from wbia import _autogen_annot_base

    BASE = _autogen_annot_base._annot_base_class
except ImportError:
    BASE = _wbia_object.ObjectList1D


# @ut.reloadable_class
@six.add_metaclass(_AnnotPropInjector)
class Annots(BASE):
    """
    Represents a group of annotations. Efficiently accesses properties from a
    database using lazy evaluation.

    CommandLine:
        python -m wbia.annots Annots

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.annots import *  # NOQA
        >>> import wbia
        >>> ibs = wbia.opendb(defaultdb='testdb1')
        >>> aids = ibs.get_valid_aids()
        >>> a = self = annots = Annots(aids, ibs)
        >>> a.preload('vecs', 'kpts', 'nids')
        >>> print(Annots.mro())
        >>> print(ut.depth_profile(a.vecs))
        >>> print(a)

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.annots import *  # NOQA
        >>> import wbia
        >>> ibs = wbia.opendb(defaultdb='testdb1')
        >>> aids = ibs.get_valid_aids()
        >>> a = self = annots = Annots(aids, ibs)
        >>> a.preload('vecs', 'kpts', 'nids')
        >>> a.disconnect()
        >>> assert 'vecs' in a._internal_attrs.keys()
        >>> assert a._ibs is None
        >>> ut.assert_raises(KeyError, a._get_num_feats)
        >>> a._ibs = ibs
        >>> assert len(a._get_num_feats()) > 0
    """

    # def __init__(self, aids, ibs, config=None, caching=False):
    #    super(Annots, self).__init__(aids, ibs, config, caching)

    @property
    def aids(self):
        return self._rowids

    def get_stats(self, **kwargs):
        self._ibs.get_annot_stats_dict(self.aids, **kwargs)

    def print_stats(self, **kwargs):
        self._ibs.print_annot_stats(self.aids, **kwargs)

    # @property
    def get_speeds(self):
        # import vtool as vt
        edges = self.get_aidpairs()
        speeds = self._ibs.get_annotpair_speeds(edges)
        # edges = vt.pdist_indicies(len(annots))
        # speeds = self._ibs.get_unflat_annots_speeds_list([self.aids])[0]
        edge_to_speed = dict(zip(edges, speeds))
        return edge_to_speed

    def get_name_image_closure(self):
        ibs = self._ibs
        aids = self.aids
        old_aids = []
        while len(old_aids) != len(aids):
            old_aids = aids
            gids = ut.unique(ibs.get_annot_gids(aids))
            other_aids = list(set(ut.flatten(ibs.get_image_aids(gids))))
            other_nids = list(set(ibs.get_annot_nids(other_aids)))
            aids = ut.flatten(ibs.get_name_aids(other_nids))
        return aids

    def group2(self, by):
        """
        self = annots
        by = annots.static_encounter
        encounters = annots.group2(annots.static_encounter)
        """
        annots_list = self.group(by)[1]
        return AnnotGroups(annots_list, self._ibs)

    def get_aidpairs(self):
        aids = self.aids
        aid_pairs = list(it.combinations(aids, 2))
        return aid_pairs

    def get_am_rowids(self, internal=True):
        """
        if `internal is True` returns am rowids only between
        annotations in this Annots object, otherwise returns
        any am rowid that contains any aid in this Annots object.
        """
        ibs = self._ibs
        if internal:
            ams = ibs.get_annotmatch_rowids_between(self.aids, self.aids)
        else:
            ams = ut.flatten(ibs.get_annotmatch_rowids_from_aid(self.aids))
        return ams

    def matches(self, internal=True):
        ams = self.get_am_rowids(internal)
        return self._ibs.matches(ams)

    def get_am_rowids_and_pairs(self):
        ibs = self._ibs
        ams = self.get_am_rowids()
        aid_pairs = ibs.get_annotmatch_aids(ams)
        # aid_pairs = self.get_aidpairs()
        # aids1 = ut.take_column(aid_pairs, 0)
        # aids2 = ut.take_column(aid_pairs, 1)
        # ams = ibs.get_annotmatch_rowid_from_undirected_superkey(aids1, aids2)
        # flags = ut.not_list(ut.flag_None_items(ams))
        # ams = ut.compress(ams, flags)
        # aid_pairs = ut.compress(aid_pairs, flags)
        return ams, aid_pairs

    def get_am_aidpairs(self):
        ibs = self._ibs
        ams = self.get_am_rowids()
        aids1 = ibs.get_annotmatch_aid1(ams)
        aids2 = ibs.get_annotmatch_aid2(ams)
        aid_pairs = list(zip(aids1, aids2))
        return aid_pairs

    @property
    def hog_img(self):
        from wbia import core_annots

        return [core_annots.make_hog_block_image(hog) for hog in self.hog_hog]

    def append_tags(self, tags):
        self._ibs.append_annot_case_tags(self._rowids, tags)

    def remove_tags(self, tags):
        self._ibs.remove_annot_case_tags(self._rowids, tags)

    def __hash__(self):
        return hash(tuple(self.aids))

    def __lt__(self, other):
        if len(self.aids) == len(other.aids):
            if len(self.aids) == 0:
                return False
            else:
                return tuple(self) < tuple(other)
        elif len(self.aids) < len(other.aids):
            return True
        else:
            return False

    def __eq__(self, other):
        if len(self.aids) == len(other.aids):
            return all(a == b for a, b in zip(self, other))
        return False

    def show(self, *args, **kwargs):
        if len(self) != 1:
            raise ValueError('Can only show one, got {}'.format(len(self)))
        from wbia.viz import viz_chip

        for aid in self:
            return viz_chip.show_chip(self._ibs, aid, *args, **kwargs)


class _AnnotGroupPropInjector(BASE_TYPE):
    def __init__(metaself, name, bases, dct):
        super(_AnnotGroupPropInjector, metaself).__init__(name, bases, dct)
        metaself.rrr = rrr

        # TODO: move to wbia object as a group call
        def _make_unflat_getter(objname, attrname):
            ibs_funcname = 'get_%s_%s' % (objname, attrname)

            def ibs_unflat_getter(self, *args, **kwargs):
                ibs_callable = getattr(self._ibs, ibs_funcname)
                rowids = self._rowids_list
                ibs = self._ibs
                return ibs.unflat_map(ibs_callable, rowids, *args, **kwargs)

            ut.set_funcname(ibs_unflat_getter, 'unflat_' + ibs_funcname)
            return ibs_unflat_getter

        for attrname in ANNOT_BASE_ATTRS:
            if hasattr(metaself, attrname):
                logger.info('Cannot inject annot group attrname = %r' % (attrname,))
                continue
            ibs_unflat_getter = _make_unflat_getter('annot', attrname)
            setattr(metaself, '_unflat_get_' + attrname, ibs_unflat_getter)
            setattr(metaself, attrname, property(ibs_unflat_getter))


@ut.reloadable_class
@six.add_metaclass(_AnnotGroupPropInjector)
class AnnotGroups(ut.NiceRepr):
    """ Effciently handle operations on multiple groups of annotations """

    def __init__(self, annots_list, ibs):
        self._ibs = ibs
        self.annots_list = annots_list
        self._rowids_list = [a._rowids for a in self.annots_list]

    def __len__(self):
        return len(self.annots_list)

    def __nice__(self):
        import numpy as np

        len_list = ut.lmap(len, self.annots_list)
        num = len(self.annots_list)
        mean = np.mean(len_list)
        std = np.std(len_list)
        if six.PY3:
            nice = '(n=%r, μ=%.1f, σ=%.1f)' % (num, mean, std)
        else:
            nice = '(n=%r, m=%.1f, s=%.1f)' % (num, mean, std)
        return nice

    def __iter__(self):
        return iter(self.annots_list)

    def __getitem__(self, index):
        return self.annots_list[index]

    @property
    def aids(self):
        return [a.aids for a in self.annots_list]

    @property
    def images(self, config=None):
        return self._ibs.images(self.gids, config)

    @property
    def match_tags(self):
        """ returns pairwise tags within the annotation group """
        ams_list = self._ibs.get_unflat_am_rowids(self.aids)
        tags = self._ibs.unflat_map(self._ibs.get_annotmatch_case_tags, ams_list)
        return tags


class _AnnotMatchPropInjector(BASE_TYPE):
    """
    Example:
        >>> # DISABLE_DOCTEST
        >>> from wbia import _wbia_object
        >>> import wbia
        >>> ibs = wbia.opendb(defaultdb='testdb1')
        >>> objname = 'annotmatch'
        >>> blacklist = []
        >>> tup = _wbia_object._find_wbia_attrs(ibs, objname, blacklist)
        >>> attrs, settable_attrs = tup
        >>> print('attrs = ' + ut.repr4(attrs))
        >>> print('settable_attrs = ' + ut.repr4(settable_attrs))
    """

    def __init__(metaself, name, bases, dct):
        super(_AnnotMatchPropInjector, metaself).__init__(name, bases, dct)
        metaself.rrr = rrr

        attrs = [
            'aid1',
            'aid2',
            'confidence',
            'count',
            'evidence_decision',
            'meta_decision',
            'posixtime_modified',
            'reviewer',
            'tag_text',
            'case_tags',
        ]
        settable_attrs = [
            'confidence',
            'count',
            'evidence_decision',
            'meta_decision',
            'posixtime_modified',
            'reviewer',
            'tag_text',
        ]

        configurable_attrs = []
        depcache_attrs = []
        aliased_attrs = {}

        objname = 'annotmatch'
        _wbia_object._inject_getter_attrs(
            metaself,
            objname,
            attrs,
            configurable_attrs,
            None,
            depcache_attrs,
            settable_attrs,
            aliased_attrs,
        )


@six.add_metaclass(_AnnotMatchPropInjector)
class AnnotMatches(BASE):
    """
    Represents a group of annotations. Efficiently accesses properties from a
    database using lazy evaluation.

    CommandLine:
        python -m wbia.annots Annots

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.annots import *  # NOQA
        >>> import wbia
        >>> ibs = wbia.opendb(defaultdb='testdb1')
        >>> aids = ibs.get_valid_aids()
        >>> annots = Annots(aids, ibs)
        >>> ams = annots.get_am_rowids()
        >>> matches = self = ibs.matches()
        >>> ed1 = matches.evidence_decision
        >>> md2 = matches.meta_decision
        >>> table = ibs.db.get_table_as_pandas('annotmatch')
        >>> assert len(table) == len(matches)
    """

    @property
    def edges(self):
        return list(zip(self.aid1, self.aid2))

    @property
    def confidence_code(self):
        INT_TO_CODE = self._ibs.const.CONFIDENCE.INT_TO_CODE
        return [INT_TO_CODE[c] for c in self.confidence]

    @property
    def meta_decision_code(self):
        INT_TO_CODE = self._ibs.const.META_DECISION.INT_TO_CODE
        return [INT_TO_CODE[c] for c in self.meta_decision]

    @property
    def evidence_decision_code(self):
        INT_TO_CODE = self._ibs.const.EVIDENCE_DECISION.INT_TO_CODE
        return [INT_TO_CODE[c] for c in self.evidence_decision]
